#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2016, Ansible, inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r'''
---
module: win_reg_stat
short_description: Get information about Windows registry keys
description:
- Like M(ansible.windows.win_file), M(ansible.windows.win_reg_stat) will return whether the key/property exists.
- It also returns the sub keys and properties of the key specified.
- If specifying a property name through I(property), it will return the information specific for that property.
options:
  path:
    description: The full registry key path including the hive to search for.
    type: str
    required: yes
    aliases: [ key ]
  name:
    description:
    - The registry property name to get information for, the return json will not include the sub_keys and properties entries for the I(key) specified.
    - Set to an empty string to target the registry key's C((Default)) property value.
    type: str
    aliases: [ entry, value, property ]
notes:
- The C(properties) return value will contain an empty string key C("") that refers to the key's C(Default) value. If
  the value has not been set then this key is not returned.
seealso:
- module: ansible.windows.win_regedit
- module: community.windows.win_regmerge
author:
- Jordan Borean (@jborean93)
'''

EXAMPLES = r'''
- name: Obtain information about a registry key using short form
  ansible.windows.win_reg_stat:
    path: HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion
  register: current_version

- name: Obtain information about a registry key property
  ansible.windows.win_reg_stat:
    path: HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion
    name: CommonFilesDir
  register: common_files_dir

- name: Obtain the registry key's (Default) property
  ansible.windows.win_reg_stat:
    path: HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion
    name: ''
  register: current_version_default
'''

RETURN = r'''
changed:
  description: Whether anything was changed.
  returned: always
  type: bool
  sample: true
exists:
  description: States whether the registry key/property exists.
  returned: success and path/property exists
  type: bool
  sample: true
properties:
  description: A dictionary containing all the properties and their values in the registry key.
  returned: success, path exists and property not specified
  type: dict
  sample: {
    "" : {
      "raw_value": "",
      "type": "REG_SZ",
      "value": ""
    },
    "binary_property" : {
      "raw_value": ["0x01", "0x16"],
      "type": "REG_BINARY",
      "value": [1, 22]
    },
    "multi_string_property" : {
      "raw_value": ["a", "b"],
      "type": "REG_MULTI_SZ",
      "value": ["a", "b"]
    }
  }
sub_keys:
  description: A list of all the sub keys of the key specified.
  returned: success, path exists and property not specified
  type: list
  sample: [
    "AppHost",
    "Casting",
    "DateTime"
  ]
raw_value:
  description: Returns the raw value of the registry property, REG_EXPAND_SZ has no string expansion, REG_BINARY or REG_NONE is in hex 0x format.
    REG_NONE, this value is a hex string in the 0x format.
  returned: success, path/property exists and property specified
  type: str
  sample: '%ProgramDir%\\Common Files'
type:
  description: The property type.
  returned: success, path/property exists and property specified
  type: str
  sample: "REG_EXPAND_SZ"
value:
  description: The value of the property.
  returned: success, path/property exists and property specified
  type: str
  sample: 'C:\\Program Files\\Common Files'
'''



#!/usr/bin/python
#
# Copyright 2016 Red Hat | Ansible
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: docker_image_remove

short_description: Remove Docker images

version_added: 3.6.0

description:
  - Remove Docker images from the Docker daemon.
extends_documentation_fragment:
  - community.docker._docker.api_documentation
  - community.docker._attributes
  - community.docker._attributes.actiongroup_docker

attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
  idempotent:
    support: full

options:
  name:
    description:
      - 'Image name. Name format will be one of: C(name), C(repository/name), C(registry_server:port/name). When pushing or
        pulling an image the name can optionally include the tag by appending C(:tag_name).'
      - Note that image IDs (hashes) can also be used.
    type: str
    required: true
  tag:
    description:
      - Tag for the image name O(name) that is to be tagged.
      - If O(name)'s format is C(name:tag), then the tag value from O(name) will take precedence.
    type: str
    default: latest
  force:
    description:
      - Un-tag and remove all images matching the specified name.
    type: bool
    default: false
  prune:
    description:
      - Delete untagged parent images.
    type: bool
    default: true

requirements:
  - "Docker API >= 1.25"

author:
  - Felix Fontein (@felixfontein)

seealso:
  - module: community.docker.docker_image_load
  - module: community.docker.docker_image_pull
  - module: community.docker.docker_image_tag
"""

EXAMPLES = r"""
---
- name: Remove an image
  community.docker.docker_image_remove:
    name: pacur/centos-7
"""

RETURN = r"""
image:
  description:
    - Image inspection results for the affected image before removal.
    - Empty if the image was not found.
  returned: success
  type: dict
  sample: {}
deleted:
  description:
    - The digests of the images that were deleted.
  returned: success
  type: list
  elements: str
  sample: []
untagged:
  description:
    - The digests of the images that were untagged.
  returned: success
  type: list
  elements: str
  sample: []
"""

import traceback
import typing as t

from ansible_collections.community.docker.plugins.module_utils._api.errors import (
    DockerException,
    NotFound,
)
from ansible_collections.community.docker.plugins.module_utils._api.utils.utils import (
    parse_repository_tag,
)
from ansible_collections.community.docker.plugins.module_utils._common_api import (
    AnsibleDockerClient,
    RequestException,
)
from ansible_collections.community.docker.plugins.module_utils._util import (
    DockerBaseClass,
    is_image_name_id,
    is_valid_tag,
)


class ImageRemover(DockerBaseClass):
    def __init__(self, client: AnsibleDockerClient) -> None:
        super().__init__()

        self.client = client
        self.check_mode = self.client.check_mode
        self.diff = self.client.module._diff

        parameters = self.client.module.params
        self.name = parameters["name"]
        self.tag = parameters["tag"]
        self.force = parameters["force"]
        self.prune = parameters["prune"]

        if not is_valid_tag(self.tag, allow_empty=True):
            self.fail(f'"{self.tag}" is not a valid docker tag')

        # If name contains a tag, it takes precedence over tag parameter.
        if not is_image_name_id(self.name):
            repo, repo_tag = parse_repository_tag(self.name)
            if repo_tag:
                self.name = repo
                self.tag = repo_tag

    def fail(self, msg: str) -> t.NoReturn:
        self.client.fail(msg)

    def get_diff_state(self, image: dict[str, t.Any] | None) -> dict[str, t.Any]:
        if not image:
            return {"exists": False}
        return {
            "exists": True,
            "id": image["Id"],
            "tags": sorted(image.get("RepoTags") or []),
            "digests": sorted(image.get("RepoDigests") or []),
        }

    def absent(self) -> dict[str, t.Any]:
        actions: list[str] = []
        deleted: list[str] = []
        untagged: list[str] = []
        results: dict[str, t.Any] = {
            "changed": False,
            "actions": actions,
            "image": {},
            "deleted": deleted,
            "untagged": untagged,
        }

        name = self.name
        if is_image_name_id(name):
            image = self.client.find_image_by_id(name, accept_missing_image=True)
        else:
            image = self.client.find_image(name, self.tag)
            if self.tag:
                name = f"{self.name}:{self.tag}"

        diff: dict[str, t.Any] = {}
        if self.diff:
            results["diff"] = diff
            diff["before"] = self.get_diff_state(image)

        if not image:
            if self.diff:
                diff["after"] = self.get_diff_state(image)
            return results

        results["changed"] = True
        actions.append(f"Removed image {name}")
        results["image"] = image

        if not self.check_mode:
            try:
                res = self.client.delete_json(
                    "/images/{0}",
                    name,
                    params={"force": self.force, "noprune": not self.prune},
                )
            except NotFound:
                # If the image vanished while we were trying to remove it, do not fail
                res = []
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.fail(f"Error removing image {name} - {exc}")

            for entry in res:
                if entry.get("Untagged"):
                    untagged.append(entry["Untagged"])
                if entry.get("Deleted"):
                    deleted.append(entry["Deleted"])

            untagged[:] = sorted(untagged)
            deleted[:] = sorted(deleted)

            if self.diff:
                image_after = self.client.find_image_by_id(
                    image["Id"], accept_missing_image=True
                )
                diff["after"] = self.get_diff_state(image_after)

        elif is_image_name_id(name):
            deleted.append(image["Id"])
            # TODO: the following is no longer correct with Docker 29+...
            untagged[:] = sorted(
                (image.get("RepoTags") or []) + (image.get("RepoDigests") or [])
            )
            if not self.force and results["untagged"]:
                self.fail(
                    "Cannot delete image by ID that is still in use - use force=true"
                )
            if self.diff:
                diff["after"] = self.get_diff_state({})

        elif is_image_name_id(self.tag):
            untagged.append(name)
            if (
                len(image.get("RepoTags") or []) < 1
                and len(image.get("RepoDigests") or []) < 2
            ):
                deleted.append(image["Id"])
            if self.diff:
                diff["after"] = self.get_diff_state(image)
                try:
                    diff["after"]["digests"].remove(name)
                except ValueError:
                    pass

        else:
            untagged.append(name)
            if (
                len(image.get("RepoTags") or []) < 2
                and len(image.get("RepoDigests") or []) < 1
            ):
                deleted.append(image["Id"])
            if self.diff:
                diff["after"] = self.get_diff_state(image)
                try:
                    diff["after"]["tags"].remove(name)
                except ValueError:
                    pass

        return results


def main() -> None:
    argument_spec = {
        "name": {"type": "str", "required": True},
        "tag": {"type": "str", "default": "latest"},
        "force": {"type": "bool", "default": False},
        "prune": {"type": "bool", "default": True},
    }

    client = AnsibleDockerClient(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    try:
        results = ImageRemover(client).absent()
        client.module.exit_json(**results)
    except DockerException as e:
        client.fail(
            f"An unexpected Docker error occurred: {e}",
            exception=traceback.format_exc(),
        )
    except RequestException as e:
        client.fail(
            f"An unexpected requests error occurred when trying to talk to the Docker daemon: {e}",
            exception=traceback.format_exc(),
        )


if __name__ == "__main__":
    main()



# Copyright (c) 2026 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import os

from ansible.errors import AnsibleActionFail
from ansible.module_utils.common.validation import check_type_bool
from ansible.plugins.action import ActionBase


class ActionModule(ActionBase):

    def run(
        self,
        tmp: str | None = None,
        task_vars: dict[str, object] | None = None,
    ) -> dict[str, object]:
        self._supports_async = True
        self._supports_check_mode = True

        if task_vars is None:
            task_vars = dict()

        super(ActionModule, self).run(tmp, task_vars)
        del tmp  # tmp no longer has any effect

        module_args = self._task.args
        config_file = module_args.get('config_file', None)
        remote_config_file = check_type_bool(module_args.get('remote_config_file', False))

        tmpdir = self._connection._shell.tmpdir
        remove_tmpdir = False
        try:

            if config_file and not remote_config_file:
                if module_args.get('config', None):
                    raise AnsibleActionFail("parameters are mutually exclusive: config, config_file")

                if self._task.async_val:
                    raise AnsibleActionFail("async operations are not supported with local config_file")

                if not tmpdir:
                    tmpdir = self._make_tmp_path()
                    remove_tmpdir = True

                full_path = self._find_needle('files', config_file)
                tmp_src = self._connection._shell.join_path(
                    tmpdir,
                    os.path.basename(full_path),
                )
                module_args['config_file'] = tmp_src
                module_args['remote_config_file'] = True

                self._transfer_file(full_path, tmp_src)

            module_result = self._execute_module(
                module_name='ansible.windows.dsc3',
                module_args=module_args,
                task_vars=task_vars,
                wrap_async=self._task.async_val,
            )
            if (
                config_file and not remote_config_file and
                'invocation' in module_result and
                'module_args' in module_result['invocation']
            ):
                # Restores the invocation back to the original state.
                module_result['invocation']['module_args']['config_file'] = config_file
                module_result['invocation']['module_args']['remote_config_file'] = remote_config_file

            return module_result

        finally:
            if tmpdir and remove_tmpdir:
                self._remove_tmp_path(tmpdir)



#
# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The iosxr_bgp_global config file.
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created.
"""


from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module import (
    ResourceModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_merge,
)

from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.facts.facts import Facts
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.rm_templates.bgp_global import (
    Bgp_globalTemplate,
)


class Bgp_global(ResourceModule):
    """
    The iosxr_bgp_global config class
    """

    def __init__(self, module):
        super(Bgp_global, self).__init__(
            empty_fact_val={},
            facts_module=Facts(module),
            module=module,
            resource="bgp_global",
            tmplt=Bgp_globalTemplate(),
        )
        self.parsers = [
            "router",
            "bfd_multiplier",
            "bfd_minimum_interval",
            "bgp_auto_policy_soft_reset",
            "bgp_as_path_loopcheck",
            "bgp_cluster_id",
            "bgp_default_local_preference",
            "bgp_enforce_first_as_disable",
            "bgp_fast_external_fallover_disable",
            "bgp_install_diversion",
            "bgp_max_neighbors",
            "bgp_redistribute_internal",
            "bgp_router_id",
            "bgp_scan_time",
            "bgp_unsafe_ebgp_policy",
            "bgp_update_delay",
            "bgp_bestpath_aigp",
            "bgp_bestpath_as_path_ignore",
            "bgp_bestpath_as_path_multipath_relax",
            "bgp_bestpath_med_always",
            "bgp_bestpath_med_confed",
            "bgp_bestpath_med_missing_as_worst",
            "bgp_bestpath_compare_routerid",
            "bgp_bestpath_cost_community_ignore",
            "bgp_bestpath_origin_as_use",
            "bgp_bestpath_origin_as_allow",
            "bgp_confederation_identifier",
            "bgp_graceful_restart_set",
            "bgp_graceful_restart_graceful_reset",
            "bgp_graceful_restart_restart_time",
            "bgp_graceful_restart_purge_time",
            "bgp_graceful_restart_stalepath_time",
            "bgp_log_message",
            "bgp_log_neighbor_changes_detail",
            "bgp_log_neighbor_changes_disable",
            "bgp_multipath_as_path_ignore_onwards",
            "bgp_origin_as_validation_disable",
            "bgp_origin_as_validation_signal_ibgp",
            "bgp_origin_as_validation_time_off",
            "bgp_origin_as_validation_time",
            "bgp_default_information_originate",
            "bgp_default_metric",
            "bgp_graceful_maintenance",
            "ibgp_policy_out_enforce_modifications",
            "mpls_activate_interface",
            "mvpn",
            "nsr_set",
            "nsr_disable",
            "socket_receive_buffer_size",
            "socket_send_buffer_size",
            "update_in_error_handling_basic_ebgp_disable",
            "update_in_error_handling_basic_ibgp_disable",
            "update_in_error_handling_extended_ebgp",
            "update_in_error_handling_extended_ibgp",
            "update_out_logging",
            "update_limit",
            "rpki_route_value",
            "rd_auto",
            "timers_keepalive",
        ]

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            # import epdb;epdb.serve()
            self.run_commands()
        return self.result

    def generate_commands(self):
        """Generate configuration commands to send based on
        want, have and desired state.
        """
        if self.state in ["merged", "replaced", "overridden"]:
            w_asn = self.want.get("as_number")
            h_asn = self.have.get("as_number")

            if h_asn and w_asn != h_asn:
                self._module.fail_json(
                    msg="BGP is already configured with ASN {0}. "
                    "Please remove it with state purged before "
                    "configuring new as_number".format(h_asn),
                )
        for entry in self.want, self.have:
            self._bgp_list_to_dict(entry)

        # if state is deleted, clean up global params
        if self.state == "deleted":
            if not self.want or (self.have.get("as_number") == self.want.get("as_number")):
                self._compare(
                    want={"as_number": self.want.get("as_number")},
                    have=self.have,
                )
        elif self.state == "purged":
            if not self.want or (self.have.get("as_number") == self.want.get("as_number")):
                self.addcmd(self.have or {}, "router", True)

        else:
            wantd = self.want
            # if state is merged, merge want onto have and then compare
            if self.state == "merged":
                wantd = dict_merge(self.have, self.want)

            self._compare(want=wantd, have=self.have)

    def _compare(self, want, have):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the Bgp_global network resource.
        """

        self.compare(parsers=self.parsers, want=want, have=have)
        self._compare_rpki_server(want=want, have=self.have)
        self._compare_confederation_peers(want=want, have=have)
        self._compare_neighbors(want=want, have=self.have)
        self._vrfs_compare(want=want, have=have)
        if self.commands and "router bgp" not in self.commands[0]:
            self.commands.insert(
                0,
                self._tmplt.render(
                    {"as_number": want["as_number"]},
                    "router",
                    False,
                ),
            )

    def _compare_rpki_server(self, want, have):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the Bgp_global rpki servers resource.
        """
        rpki_server_parsers = [
            "rpki_server_purge_time",
            "rpki_server_refresh_time",
            "rpki_server_refresh_time_off",
            "rpki_server_response_time",
            "rpki_server_response_time_off",
            "rpki_server_shutdown",
            "rpki_server_transport_ssh",
            "rpki_server_transport_tcp",
        ]
        want = want.get("rpki", {}).get("servers", {})
        have = have.get("rpki", {}).get("servers", {})
        for name, entry in want.items():
            new_have = have.pop(name, {})
            begin = len(self.commands)
            self.compare(
                parsers=rpki_server_parsers,
                want=entry,
                have=new_have,
            )
            rpki_server_name = entry.get("name")
            if len(self.commands) != begin:
                self.commands.insert(
                    begin,
                    self._tmplt.render(
                        {"name": rpki_server_name},
                        "rpki_server_name",
                        False,
                    ),
                )
        for name, entry in have.items():
            self.addcmd(entry, "rpki_server_name", True)

    def _compare_confederation_peers(self, want, have):
        """Custom handling of confederation.peers option
        :params want: the want BGP dictionary
        :params have: the have BGP dictionary
        """
        w_cpeers = want.get("bgp", {}).get("confederation", {}).get("peers", [])
        h_cpeers = have.get("bgp", {}).get("confederation", {}).get("peers", [])

        if set(w_cpeers) != set(h_cpeers):
            if self.state in ["replaced", "deleted"]:
                # if there are peers already configured
                # we need to remove those before we pass
                # the new ones otherwise the device appends
                # them to the existing ones
                if h_cpeers:
                    self.addcmd(have, "bgp_confederation_peers", True)
            if w_cpeers:
                self.addcmd(want, "bgp_confederation_peers", False)

    def _compare_neighbors(self, want, have, vrf=None):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the Bgp_global neighbor resource.
        """
        neighbor_parsers = [
            "use.neighbor_group",
            "use.session_group",
            "advertisement_interval",
            "bfd_fast_detect_disable",
            "bfd_fast_detect_strict_mode",
            "bfd_fast_detect_set",
            "bfd_nbr_minimum_interval",
            "bfd_nbr_multiplier",
            "bmp_activate",
            "dmz_link_bandwidth",
            "dmz_link_bandwidth_inheritance_disable",
            "neighbor_description",
            "neighbor_cluster_id",
            "dscp",
            "ebgp_multihop_value",
            "ebgp_multihop_mpls",
            "ebgp_recv_extcommunity_dmz",
            "ebgp_recv_extcommunity_dmz_set",
            "ebgp_send_extcommunity_dmz",
            "ebgp_send_extcommunity_dmz_set",
            "ebgp_send_extcommunity_dmz_cumulatie",
            "egress_engineering",
            "egress_engineering_set",
            "ignore_connected_check",
            "ignore_connected_check_set",
            "neighbor_enforce_first_as_disable",
            "neighbor_graceful_restart_restart_time",
            "neighbor_graceful_restart_stalepath_time",
            "keychain",
            "keychain_name",
            "local_as_inheritance_disable",
            "local_as",
            "local",
            "local_address",
            "origin_as",
            "password_inheritance_disable",
            "password_encrypted",
            "remote_as",
            "receive_buffer_size",
            "send_buffer_size",
            "session_open_mode",
            "neighbor_shutdown",
            "neighbor_shutdown_inheritance_disable",
            "neighbor_tcp_mss",
            "neighbor_tcp_mss_inheritance_disable",
            "neighbor_timers_keepalive",
            "update_source",
            "neighbor_ttl_security_inheritance_disable",
            "neighbor_ttl_security",
            "neighbor_graceful_maintenance_set",
            "neighbor_graceful_maintenance_activate",
            "neighbor_graceful_maintenance_activate_inheritance_disable",
            "neighbor_graceful_maintenance_as_prepends",
            "neighbor_graceful_maintenance_local_preference_disable",
            "neighbor_graceful_maintenance_local_preference",
            "neighbor_graceful_maintenance_as_prepends_value",
            "neighbor_capability_additional_paths_send",
            "neighbor_capability_additional_paths_send_disable",
            "neighbor_capability_additional_paths_rcv_disable",
            "neighbor_capability_additional_paths_rcv",
            "neighbor_capability_suppress_four_byte_AS",
            "neighbor_capability_suppress_all",
            "neighbor_capability_suppress_all_inheritance_disable",
            "neighbor_log_message_in_value",
            "neighbor_log_message_in_disable",
            "neighbor_log_message_in_inheritance_disable",
            "neighbor_log_message_out_value",
            "neighbor_log_message_out_disable",
            "neighbor_log_message_out_inheritance_disable",
            "neighbor_update_in_filtering_attribute_filter_group",
            "neighbor_update_in_filtering_logging_disable",
            "neighbor_update_in_filtering_message_buffers",
        ]

        want_nbr = want.get("neighbors", {})
        have_nbr = have.get("neighbors", {})
        for name, entry in want_nbr.items():
            have = have_nbr.pop(name, {})
            begin = len(self.commands)
            self.compare(parsers=neighbor_parsers, want=entry, have=have)
            neighbor_address = entry.get("neighbor_address", "")
            if len(self.commands) != begin:
                self.commands.insert(
                    begin,
                    self._tmplt.render(
                        {"neighbor_address": neighbor_address},
                        "neighbor_address",
                        False,
                    ),
                )
        for name, entry in have_nbr.items():
            if self._check_af("neighbor_address", name):
                self._module.fail_json(
                    msg="Neighbor {0} has address-family configurations. "
                    "Please use the iosxr_bgp_neighbor_address_family module to remove those first.".format(
                        name,
                    ),
                )
            else:
                self.addcmd(entry, "neighbor_address", True)

    def _vrfs_compare(self, want, have):
        """Custom handling of VRFs option
        :params want: the want BGP dictionary
        :params have: the have BGP dictionary
        """
        wvrfs = want.get("vrfs", {})
        hvrfs = have.get("vrfs", {})
        for name, entry in wvrfs.items():
            begin = len(self.commands)
            vrf_have = hvrfs.pop(name, {})
            self._compare_rpki_server(want=entry, have=vrf_have)
            self._compare_neighbors(want=entry, have=vrf_have)
            self.compare(parsers=self.parsers, want=entry, have=vrf_have)
            if len(self.commands) != begin:
                self.commands.insert(
                    begin,
                    self._tmplt.render(
                        {"vrf": entry.get("vrf")},
                        "vrf",
                        False,
                    ),
                )
        # cleanup remaining VRFs
        # but do not negate it entirely
        # instead remove only those attributes
        # that this module manages
        for name, entry in hvrfs.items():
            if self._check_af("vrf", name):
                self._module.fail_json(
                    msg="VRF {0} has address-family configurations. "
                    "Please use the iosxr_bgp_address_family module to remove those first.".format(
                        name,
                    ),
                )
            else:
                self.addcmd(entry, "vrf", True)

    def _bgp_list_to_dict(self, entry):
        """Convert list of items to dict of items
           for efficient diff calculation.
        :params entry: data dictionary
        """

        def _build_key(x):
            """Build primary key for path_attribute
               option.
            :params x: path_attribute dictionary
            :returns: primary key as tuple
            """
            key_1 = "start_{0}".format(x.get("range", {}).get("start", ""))
            key_2 = "end_{0}".format(x.get("range", {}).get("end", ""))
            key_3 = "type_{0}".format(x.get("type", ""))
            key_4 = x["action"]

            return (key_1, key_2, key_3, key_4)

        if "servers" in entry.get("rpki", {}):
            entry["rpki"]["servers"] = {
                x["name"]: x for x in entry.get("rpki", {}).get("servers", [])
            }
        if "neighbors" in entry:
            entry["neighbors"] = {x["neighbor_address"]: x for x in entry.get("neighbors", [])}

        if "vrfs" in entry:
            entry["vrfs"] = {x["vrf"]: x for x in entry.get("vrfs", [])}
            for _k, vrf in entry["vrfs"].items():
                self._bgp_list_to_dict(vrf)

    def _get_config(self):
        return self._connection.get("show running-config router bgp")

    def _check_af(self, context, context_name):
        af_present = False
        if self._connection:
            config_lines = self._get_config().splitlines()
            index = [i + 1 for i, el in enumerate(config_lines) if context_name in el]
            if index:
                # had to do this to escape flake8 and black errors
                ind = index[0]
                for line in config_lines[ind:]:
                    if context in line:
                        break
                    if "address-family" in line:
                        af_present = True
                        break
        return af_present



# Copyright (c) 2018, David Kainz <dkainz@mgit.at> <dave.jokain@gmx.at>
# Copyright (c) 2021, Andrew Pantuso (@ajpantuso) <ajpantuso@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import abc
import os
import typing as t

from ansible.module_utils.basic import missing_required_lib
from ansible.module_utils.common.text.converters import to_bytes, to_text

from ansible_collections.community.crypto.plugins.module_utils._cryptography_dep import (
    COLLECTION_MINIMUM_CRYPTOGRAPHY_VERSION,
)
from ansible_collections.community.crypto.plugins.module_utils._openssh.backends.common import (
    KeygenCommand,
    OpensshModule,
    PrivateKey,
    PublicKey,
    parse_private_key_format,
)
from ansible_collections.community.crypto.plugins.module_utils._openssh.cryptography import (
    CRYPTOGRAPHY_VERSION,
    HAS_OPENSSH_SUPPORT,
    InvalidCommentError,
    InvalidPassphraseError,
    InvalidPrivateKeyFileError,
    OpenSSHError,
    OpensshKeypair,
)
from ansible_collections.community.crypto.plugins.module_utils._openssh.utils import (
    any_in,
    file_mode,
    secure_write,
)
from ansible_collections.community.crypto.plugins.module_utils._version import (
    LooseVersion,
)

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule  # pragma: no cover


class KeypairBackend(OpensshModule, metaclass=abc.ABCMeta):
    def __init__(self, *, module: AnsibleModule) -> None:
        super().__init__(module=module)

        self.comment: str | None = self.module.params["comment"]
        self.private_key_path: str = self.module.params["path"]
        self.public_key_path = self.private_key_path + ".pub"
        self.regenerate: t.Literal[
            "never", "fail", "partial_idempotence", "full_idempotence", "always"
        ] = (
            self.module.params["regenerate"]
            if not self.module.params["force"]
            else "always"
        )
        self.state: t.Literal["present", "absent"] = self.module.params["state"]
        self.type: t.Literal["rsa", "dsa", "rsa1", "ecdsa", "ed25519"] = (
            self.module.params["type"]
        )

        self.size: int = self._get_size(self.module.params["size"])
        self._validate_path()

        self.original_private_key: PrivateKey | None = None
        self.original_public_key: PublicKey | None = None
        self.private_key: PrivateKey | None = None
        self.public_key: PublicKey | None = None

    def _get_size(self, size: int | None) -> int:
        if self.type in ("rsa", "rsa1"):
            result = 4096 if size is None else size
            if result < 1024:
                return self.module.fail_json(
                    msg="For RSA keys, the minimum size is 1024 bits and the default is 4096 bits. "
                    + "Attempting to use bit lengths under 1024 will cause the module to fail."
                )
        elif self.type == "dsa":
            result = 1024 if size is None else size
            if result != 1024:
                return self.module.fail_json(
                    msg="DSA keys must be exactly 1024 bits as specified by FIPS 186-2."
                )
        elif self.type == "ecdsa":
            result = 256 if size is None else size
            if result not in (256, 384, 521):
                return self.module.fail_json(
                    msg="For ECDSA keys, size determines the key length by selecting from one of "
                    + "three elliptic curve sizes: 256, 384 or 521 bits. "
                    + "Attempting to use bit lengths other than these three values for ECDSA keys will "
                    + "cause this module to fail."
                )
        elif self.type == "ed25519":
            # User input is ignored for `key size` when `key type` is ed25519
            result = 256
        else:
            return self.module.fail_json(
                msg=f"{self.type} is not a valid value for key type"
            )

        return result

    def _validate_path(self) -> None:
        self._check_if_base_dir(self.private_key_path)

        if os.path.isdir(self.private_key_path):
            self.module.fail_json(
                msg=f"{self.private_key_path} is a directory. Please specify a path to a file."
            )

    def _execute(self) -> None:
        self.original_private_key = self._load_private_key()
        self.original_public_key = self._load_public_key()

        if self.state == "present":
            self._validate_key_load()

            if self._should_generate():
                self._generate()
            elif not self._public_key_valid():
                self._restore_public_key()

            self.private_key = self._load_private_key()
            self.public_key = self._load_public_key()

            for path in (self.private_key_path, self.public_key_path):
                self._update_permissions(path)
        else:
            if self._should_remove():
                self._remove()

    def _load_private_key(self) -> PrivateKey | None:
        result = None
        if self._private_key_exists():
            try:
                result = self._get_private_key()
            except Exception:
                pass

        return result

    def _private_key_exists(self) -> bool:
        return os.path.exists(self.private_key_path)

    @abc.abstractmethod
    def _get_private_key(self) -> PrivateKey:
        pass

    def _load_public_key(self) -> PublicKey | None:
        result = None
        if self._public_key_exists():
            try:
                result = PublicKey.load(self.public_key_path)
            except (IOError, OSError):
                pass
        return result

    def _public_key_exists(self) -> bool:
        return os.path.exists(self.public_key_path)

    def _validate_key_load(self) -> None:
        if (
            self._private_key_exists()
            and self.regenerate in ("never", "fail", "partial_idempotence")
            and (self.original_private_key is None or not self._private_key_readable())
        ):
            self.module.fail_json(
                msg="Unable to read the key. The key is protected with a passphrase or broken. "
                + "Will not proceed. To force regeneration, call the module with `generate` "
                + "set to `full_idempotence` or `always`, or with `force=true`."
            )

    @abc.abstractmethod
    def _private_key_readable(self) -> bool:
        pass

    def _should_generate(self) -> bool:
        if self.original_private_key is None:
            return True
        if self.regenerate == "never":
            return False
        if self.regenerate == "fail":
            if not self._private_key_valid():
                self.module.fail_json(
                    msg="Key has wrong type and/or size. Will not proceed. "
                    + "To force regeneration, call the module with `generate` set to "
                    + "`partial_idempotence`, `full_idempotence` or `always`, or with `force=true`."
                )
            return False
        if self.regenerate in ("partial_idempotence", "full_idempotence"):
            return not self._private_key_valid()
        return True

    def _private_key_valid(self) -> bool:
        if self.original_private_key is None:
            return False

        return all(
            [
                self.size == self.original_private_key.size,
                self.type == self.original_private_key.type,
                self._private_key_valid_backend(self.original_private_key),
            ]
        )

    @abc.abstractmethod
    def _private_key_valid_backend(self, original_private_key: PrivateKey) -> bool:
        pass

    @OpensshModule.trigger_change
    @OpensshModule.skip_if_check_mode
    def _generate(self) -> None:
        temp_private_key, temp_public_key = self._generate_temp_keypair()

        try:
            self._safe_secure_move(
                [
                    (temp_private_key, self.private_key_path),
                    (temp_public_key, self.public_key_path),
                ]
            )
        except OSError as e:
            self.module.fail_json(msg=str(e))

    def _generate_temp_keypair(self) -> tuple[str, str]:
        temp_private_key = os.path.join(
            self.module.tmpdir, os.path.basename(self.private_key_path)
        )
        temp_public_key = temp_private_key + ".pub"

        try:
            self._generate_keypair(temp_private_key)
        except (IOError, OSError) as e:
            self.module.fail_json(msg=str(e))

        for f in (temp_private_key, temp_public_key):
            self.module.add_cleanup_file(f)

        return temp_private_key, temp_public_key

    @abc.abstractmethod
    def _generate_keypair(self, private_key_path: str) -> None:
        pass

    def _public_key_valid(self) -> bool:
        if self.original_public_key is None:
            return False

        valid_public_key = self._get_public_key()
        if valid_public_key:
            valid_public_key.comment = self.comment

        return self.original_public_key == valid_public_key

    @abc.abstractmethod
    def _get_public_key(self) -> PublicKey | t.Literal[""]:
        pass

    @OpensshModule.trigger_change
    @OpensshModule.skip_if_check_mode
    def _restore_public_key(self) -> None:
        try:
            temp_public_key = self._create_temp_public_key(
                str(self._get_public_key()) + "\n"
            )
            self._safe_secure_move([(temp_public_key, self.public_key_path)])
        except (IOError, OSError):
            self.module.fail_json(
                msg="The public key is missing or does not match the private key. "
                + "Unable to regenerate the public key."
            )

        if self.comment:
            self._update_comment()

    def _create_temp_public_key(self, content: str | bytes) -> str:
        temp_public_key: str = os.path.join(
            self.module.tmpdir, os.path.basename(self.public_key_path)
        )

        default_permissions = 0o644
        existing_permissions = file_mode(self.public_key_path)

        try:
            secure_write(
                path=temp_public_key,
                mode=existing_permissions or default_permissions,
                content=to_bytes(content),
            )
        except (IOError, OSError) as e:
            self.module.fail_json(msg=str(e))
        self.module.add_cleanup_file(temp_public_key)

        return temp_public_key

    @abc.abstractmethod
    def _update_comment(self) -> None:
        pass

    def _should_remove(self) -> bool:
        return self._private_key_exists() or self._public_key_exists()

    @OpensshModule.trigger_change
    @OpensshModule.skip_if_check_mode
    def _remove(self) -> None:
        try:
            if self._private_key_exists():
                os.remove(self.private_key_path)
            if self._public_key_exists():
                os.remove(self.public_key_path)
        except (IOError, OSError) as e:
            self.module.fail_json(msg=str(e))

    @property
    def _result(self) -> dict[str, t.Any]:
        private_key = self.private_key or self.original_private_key
        public_key = self.public_key or self.original_public_key

        return {
            "size": self.size,
            "type": self.type,
            "filename": self.private_key_path,
            "fingerprint": private_key.fingerprint if private_key else "",
            "public_key": str(public_key) if public_key else "",
            "comment": public_key.comment if public_key else "",
        }

    @property
    def diff(self) -> dict[str, t.Any]:
        before = (
            self.original_private_key.to_dict() if self.original_private_key else {}
        )
        before.update(
            self.original_public_key.to_dict() if self.original_public_key else {}
        )

        after = self.private_key.to_dict() if self.private_key else {}
        after.update(self.public_key.to_dict() if self.public_key else {})

        return {
            "before": before,
            "after": after,
        }


class KeypairBackendOpensshBin(KeypairBackend):
    def __init__(self, *, module: AnsibleModule) -> None:
        super().__init__(module=module)

        if self.module.params["private_key_format"] != "auto":
            self.module.fail_json(
                msg="'auto' is the only valid option for 'private_key_format' when 'backend' is not 'cryptography'"
            )

        self.ssh_keygen = KeygenCommand(self.module)

    def _generate_keypair(self, private_key_path: str) -> None:
        self.ssh_keygen.generate_keypair(
            private_key_path=private_key_path,
            size=self.size,
            key_type=self.type,
            comment=self.comment,
            check_rc=True,
        )

    def _get_private_key(self) -> PrivateKey:
        rc, private_key_content, err = self.ssh_keygen.get_private_key(
            private_key_path=self.private_key_path, check_rc=False
        )
        if rc != 0:
            raise ValueError(err)
        return PrivateKey.from_string(private_key_content)

    def _get_public_key(self) -> PublicKey | t.Literal[""]:
        public_key_content = self.ssh_keygen.get_matching_public_key(
            private_key_path=self.private_key_path, check_rc=True
        )[1]
        return PublicKey.from_string(public_key_content)

    def _private_key_readable(self) -> bool:
        rc, _stdout, stderr = self.ssh_keygen.get_matching_public_key(
            private_key_path=self.private_key_path, check_rc=False
        )
        return not (
            rc == 255
            or any_in(
                stderr,
                "is not a public key file",
                "incorrect passphrase",
                "load failed",
            )
        )

    def _update_comment(self) -> None:
        assert self.comment is not None
        try:
            ssh_version = self._get_ssh_version() or "7.8"
            force_new_format = (
                LooseVersion("6.5") <= LooseVersion(ssh_version) < LooseVersion("7.8")
            )
            self.ssh_keygen.update_comment(
                private_key_path=self.private_key_path,
                comment=self.comment or "",
                force_new_format=force_new_format,
                check_rc=True,
            )
        except (IOError, OSError) as e:
            self.module.fail_json(msg=str(e))

    def _private_key_valid_backend(self, original_private_key: PrivateKey) -> bool:
        return True


class KeypairBackendCryptography(KeypairBackend):
    def __init__(self, *, module: AnsibleModule) -> None:
        super().__init__(module=module)

        if self.type == "rsa1":
            self.module.fail_json(
                msg="RSA1 keys are not supported by the cryptography backend"
            )

        self.passphrase = (
            to_bytes(module.params["passphrase"])
            if module.params["passphrase"]
            else None
        )
        key_format: t.Literal["auto", "pkcs1", "pkcs8", "ssh"] = module.params[
            "private_key_format"
        ]
        self.private_key_format = self._get_key_format(key_format)

    def _get_key_format(
        self, key_format: t.Literal["auto", "pkcs1", "pkcs8", "ssh"]
    ) -> t.Literal["SSH", "PKCS1", "PKCS8"]:
        result: t.Literal["SSH", "PKCS1", "PKCS8"] = "SSH"

        if key_format == "auto":
            # Default to OpenSSH 7.8 compatibility when OpenSSH is not installed
            ssh_version = self._get_ssh_version() or "7.8"

            if (
                LooseVersion(ssh_version) < LooseVersion("7.8")
                and self.type != "ed25519"
            ):
                # OpenSSH made SSH formatted private keys available in version 6.5,
                # but still defaulted to PKCS1 format with the exception of ed25519 keys
                result = "PKCS1"
        else:
            result = key_format.upper()  # type: ignore

        return result

    def _generate_keypair(self, private_key_path: str) -> None:
        assert self.type != "rsa1"
        keypair = OpensshKeypair.generate(
            keytype=self.type,
            size=self.size,
            passphrase=self.passphrase,
            comment=self.comment or "",
        )

        encoded_private_key = OpensshKeypair.encode_openssh_privatekey(
            asym_keypair=keypair.asymmetric_keypair, key_format=self.private_key_format
        )
        secure_write(path=private_key_path, mode=0o600, content=encoded_private_key)

        public_key_path = private_key_path + ".pub"
        secure_write(path=public_key_path, mode=0o644, content=keypair.public_key)

    def _get_private_key(self) -> PrivateKey:
        keypair = OpensshKeypair.load(
            path=self.private_key_path, passphrase=self.passphrase, no_public_key=True
        )

        return PrivateKey(
            size=keypair.size,
            key_type=keypair.key_type,
            fingerprint=keypair.fingerprint,
            key_format=parse_private_key_format(path=self.private_key_path),
        )

    def _get_public_key(self) -> PublicKey | t.Literal[""]:
        try:
            keypair = OpensshKeypair.load(
                path=self.private_key_path,
                passphrase=self.passphrase,
                no_public_key=True,
            )
        except OpenSSHError:
            # Simulates the null output of ssh-keygen
            return ""

        return PublicKey.from_string(to_text(keypair.public_key))

    def _private_key_readable(self) -> bool:
        try:
            OpensshKeypair.load(
                path=self.private_key_path,
                passphrase=self.passphrase,
                no_public_key=True,
            )
        except (InvalidPrivateKeyFileError, InvalidPassphraseError):
            return False

        # Cryptography >= 3.0 uses a SSH key loader which does not raise an exception when a passphrase is provided
        # when loading an unencrypted key
        if self.passphrase:
            try:
                OpensshKeypair.load(
                    path=self.private_key_path, passphrase=None, no_public_key=True
                )
                return False
            except (InvalidPrivateKeyFileError, InvalidPassphraseError):
                return True

        return True

    def _update_comment(self) -> None:
        assert self.comment is not None
        keypair = OpensshKeypair.load(
            path=self.private_key_path, passphrase=self.passphrase, no_public_key=True
        )
        try:
            keypair.comment = self.comment
        except InvalidCommentError as e:
            self.module.fail_json(msg=str(e))

        try:
            temp_public_key = self._create_temp_public_key(keypair.public_key + b"\n")
            self._safe_secure_move([(temp_public_key, self.public_key_path)])
        except (IOError, OSError) as e:
            self.module.fail_json(msg=str(e))

    def _private_key_valid_backend(self, original_private_key: PrivateKey) -> bool:
        # avoids breaking behavior and prevents
        # automatic conversions with OpenSSH upgrades
        if self.module.params["private_key_format"] == "auto":
            return True

        return self.private_key_format == original_private_key.format


def select_backend(
    *, module: AnsibleModule, backend: t.Literal["auto", "opensshbin", "cryptography"]
) -> KeypairBackend:
    can_use_cryptography = HAS_OPENSSH_SUPPORT and LooseVersion(
        CRYPTOGRAPHY_VERSION
    ) >= LooseVersion(COLLECTION_MINIMUM_CRYPTOGRAPHY_VERSION)
    can_use_opensshbin = bool(module.get_bin_path("ssh-keygen"))

    if backend == "auto":
        if can_use_opensshbin and not module.params["passphrase"]:
            backend = "opensshbin"
        elif can_use_cryptography:
            backend = "cryptography"
        else:
            module.fail_json(
                msg=(
                    f"Cannot find either the OpenSSH binary in the PATH or cryptography >= {COLLECTION_MINIMUM_CRYPTOGRAPHY_VERSION} installed on this system"
                )
            )

    if backend == "opensshbin":
        if not can_use_opensshbin:
            module.fail_json(msg="Cannot find the OpenSSH binary in the PATH")
        return KeypairBackendOpensshBin(module=module)
    if backend == "cryptography":
        if not can_use_cryptography:
            module.fail_json(
                msg=missing_required_lib(
                    f"cryptography >= {COLLECTION_MINIMUM_CRYPTOGRAPHY_VERSION}"
                )
            )
        return KeypairBackendCryptography(module=module)
    raise ValueError(f"Unsupported value for backend: {backend}")


__all__ = ("KeypairBackend", "select_backend")



#
# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The iosxr_snmp_server config file.
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created.
"""

from copy import deepcopy

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module import (
    ResourceModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_diff,
    dict_merge,
)

from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.facts.facts import Facts
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.rm_templates.snmp_server import (
    Snmp_serverTemplate,
)


class Snmp_server(ResourceModule):
    """
    The iosxr_snmp_server config class
    """

    def __init__(self, module):
        super(Snmp_server, self).__init__(
            empty_fact_val={},
            facts_module=Facts(module),
            module=module,
            resource="snmp_server",
            tmplt=Snmp_serverTemplate(),
        )
        self.parsers = [
            "chassis_id",
            "correlator.buffer_size",
            "contact",
            "ifindex",
            "ipv4.dscp",
            "ipv6.dscp",
            "ipv4.precedence",
            "ipv6.precedence",
            "location",
            "logging_threshold_oid_processing",
            "logging_threshold_pdu_processing",
            "mib_bulkstat_max_procmem_size",
            "mroutemib_send_all_vrf",
            "oid_poll_stats",
            "overload_control",
            "packetsize",
            "queue_length",
            "throttle_time",
            "trap_source",
            "trap_timeout",
            "drop.report_IPv4",
            "drop.report_IPv6",
            "drop.unknown_user",
            "ifmib.internal_cache_max_duration",
            "ifmib.ipsubscriber",
            "ifmib.stats",
            "ifmib.ifalias_long",
            "inform.timeout",
            "inform.retries",
            "inform.pending",
            "notification_log_mib.size",
            "notification_log_mib.default",
            "notification_log_mib.disable",
            "notification_log_mib.GlobalSize",
            "trap.link_ietf",
            "trap.authentication_vrf_disable",
            "trap.throttle_time",
            "timeouts.threshold",
            "timeouts.pdu_stats",
            "timeouts.subagent",
            "timeouts.inQdrop",
            "timeouts.duplicate",
            "traps.addrpool.low",
            "traps.addrpool.high",
            "traps.bfd",
            "traps.bgp.cbgp2",
            "traps.bgp.updown",
            "traps.bulkstat_collection",
            "traps.bulkstat_transfer",
            "traps.bridgemib",
            "traps.copy_complete",
            "traps.cisco_entity_ext",
            "traps.config",
            "traps.diameter.peerdown",
            "traps.diameter.peerup",
            "traps.diameter.protocolerror",
            "traps.diameter.permanentfail",
            "traps.diameter.transientfail",
            "traps.entity",
            "traps.entity_redundancy.all",
            "traps.entity_redundancy.status",
            "traps.entity_redundancy.switchover",
            "traps.entity_state.operstatus",
            "traps.entity_state.switchover",
            "traps.flash.removal",
            "traps.flash.insertion",
            "traps.fru_ctrl",
            "traps.hsrp",
            "traps.ipsla",
            "traps.ipsec.start",
            "traps.ipsec.stop",
            "traps.isakmp.start",
            "traps.isakmp.stop",
            "traps.isis",
            "traps.l2tun.pseudowire_status",
            "traps.l2tun.sessions",
            "traps.l2tun.tunnel_up",
            "traps.l2tun.tunnel_down",
            "traps.l2vpn.all",
            "traps.l2vpn.cisco",
            "traps.l2vpn.vc_up",
            "traps.l2vpn.vc_down",
            "traps.msdp_peer_state_change",
            "traps.ospf.retransmit.virt_packets",
            "traps.ospf.retransmit.packets",
            "traps.ospf.lsa.lsa_originate",
            "traps.ospf.lsa.lsa_maxage",
            "traps.ospf.errors.bad_packet",
            "traps.ospf.errors.authentication_failure",
            "traps.ospf.errors.config_error",
            "traps.ospf.errors.virt_bad_packet",
            "traps.ospf.errors.virt_authentication_failure",
            "traps.ospf.errors.virt_config_error",
            "traps.ospf.state_change.if_state_change",
            "traps.ospf.state_change.neighbor_state_change",
            "traps.ospf.state_change.virtif_state_change",
            "traps.ospf.state_change.virtneighbor_state_change",
            "traps.ospfv3.errors.bad_packet",
            "traps.ospfv3.errors.authentication_failure",
            "traps.ospfv3.errors.config_error",
            "traps.ospfv3.errors.virt_config_error",
            "traps.ospfv3.errors.virt_bad_packet",
            "traps.ospfv3.state_change.neighbor_state_change",
            "traps.ospfv3.state_change.virtif_state_change",
            "traps.ospfv3.state_change.virtneighbor_state_change",
            "traps.ospfv3.state_change.restart_status_change",
            "traps.ospfv3.state_change.restart_helper_status_change",
            "traps.ospfv3.state_change.restart_virtual_helper_status_change",
            "traps.ospfv3.state_change.nssa_state_change",
            "traps.power",
            "traps.rf",
            "traps.pim.neighbor_change",
            "traps.pim.invalid_message_received",
            "traps.pim.rp_mapping_change",
            "traps.pim.interface_state_change",
            "traps.rsvp.lost_flow",
            "traps.rsvp.new_flow",
            "traps.rsvp.all",
            "traps.selective_vrf_download_role_change",
            "traps.sensor",
            "traps.vrrp_events",
            "traps.syslog",
            "traps.system",
            "traps.subscriber.session_agg_access_interface",
            "traps.subscriber.session_agg_node",
            "traps.vpls.all",
            "traps.vpls.full_clear",
            "traps.vpls.full_raise",
            "traps.vpls.status",
            "traps.snmp.linkup",
            "traps.snmp.linkdown",
            "traps.snmp.coldstart",
            "traps.snmp.warmstart",
            "traps.snmp.authentication",
        ]

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            self.run_commands()
        return self.result

    def handle_alieses(self, want):
        for x in [want.get("groups", []), want.get("users", [])]:
            for y in x:
                if y.get("Ipv4_acl"):
                    del y["Ipv4_acl"]
                if y.get("Ipv6_acl"):
                    del y["Ipv6_acl"]
        return want

    def generate_commands(self):
        """Generate configuration commands to send based on
        want, have and desired state.
        """

        self.want = self.handle_alieses(self.want)
        wantd = self.list_to_dict(self.want)
        haved = self.list_to_dict(self.have)

        # if state is merged, merge want onto have and then compare
        if self.state == "merged":
            wantd = dict_merge(haved, wantd)

        # if state is deleted, empty out wantd and set haved to wantd
        if self.state == "deleted":
            wantd = {}

        self._compare(want=wantd, have=haved)

        # --- ensure explicit 'no' commands for interfaces & vrfs when deleting ---
        # When state is deleted, the comparison may already produce many 'no ...' commands,
        # but ensure we always include explicit removals for interfaces and vrfs that
        # existed in `have` but weren't generated by compare.
        if self.state == "deleted":
            # ensure interfaces get explicit 'no snmp-server interface <name>'
            interfaces = haved.get("interfaces", {}) or {}
            for if_name in interfaces.keys():
                no_cmd = f"no snmp-server interface {if_name}"
                if no_cmd not in self.commands:
                    self.commands.append(no_cmd)

            # ensure vrfs get explicit 'no snmp-server vrf <name>'
            vrfs = haved.get("vrfs", {}) or {}
            for vrf_name in vrfs.keys():
                no_cmd = f"no snmp-server vrf {vrf_name}"
                if no_cmd not in self.commands:
                    self.commands.append(no_cmd)

        if self.state in ["overridden", "replaced"]:
            self.commands = [each for each in self.commands if "no" in each] + [
                each for each in self.commands if "no" not in each
            ]

    def _compare(self, want, have):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the Logging_global network resource.
        """
        self.compare(parsers=self.parsers, want=want, have=have)
        self._compare_lists(want, have)
        self._compare_vrfs(want, have)

    def _remove_snmp_server(self, begin):
        for i in range(begin, len(self.commands)):
            self.commands[i] = self.commands[i].replace("snmp-server ", "")

    def _compare_vrfs(self, want, have):
        wvrfs = want.get("vrfs", {})
        hvrfs = have.get("vrfs", {})
        for name, entry in wvrfs.items():
            begin = len(self.commands)
            vrf_have = hvrfs.pop(name, {})
            self._compare_lists(want=entry, have=vrf_have)
            if len(self.commands) != begin:
                self._remove_snmp_server(begin)
                self.commands.insert(
                    begin,
                    self._tmplt.render(
                        {"vrf": entry.get("vrf")},
                        "vrfs",
                        False,
                    ),
                )
        for name, entry in hvrfs.items():
            self.addcmd(entry, "vrfs", True)

    def _compare_lists(self, want, have):
        """
        Handles list attributes from config_data
        """
        for x in [
            "communities",
            "community_maps",
            "correlator.rule_sets",
            "correlator.rules",
            "context",
            "groups",
            "hosts",
            "interfaces",
            "mib_object_lists",
            "mib_schema",
            "mib_bulkstat_transfer_ids",
            "users",
            "targets",
        ]:
            wantx = want.get(x, {})
            havex = have.get(x, {})
            if "." in x:
                complex_parser = x.split(".")
                wantx = want.get(complex_parser[0], {}).get(
                    complex_parser[1],
                    {},
                )
                havex = have.get(complex_parser[0], {}).get(
                    complex_parser[1],
                    {},
                )

            if x in [
                "interfaces",
                "correlator.rules",
                "mib_schema",
                "mib_bulkstat_transfer_ids",
            ]:
                # handling complex parsers for replaced and overridden state

                for key, wentry in wantx.items():
                    hentry = havex.pop(key, {})
                    updates = dict_diff(hentry, wentry)
                    if updates and x in [
                        "interfaces",
                        "mib_schema",
                        "mib_bulkstat_transfer_ids",
                    ]:
                        updates.update(name=wentry["name"])
                        self.addcmd(updates, x)
                    elif updates and x == "correlator.rules":
                        updates.update(rule_name=wentry["rule_name"])
                        self.addcmd(updates, x)
            else:
                for key, wentry in wantx.items():
                    hentry = havex.pop(key, {})
                    if wentry != hentry:
                        self.addcmd(wentry, x)

                for key, hentry in havex.items():
                    self.addcmd(hentry, x, negate=True)

    def _host_list_to_dict(self, data):
        host_dict = {}
        host_data = deepcopy(data)
        for el in host_data["hosts"]:
            tr = ""
            inf = ""
            if el.get("traps"):
                tr = "traps"
            if el.get("informs"):
                inf = "informs"
            host_dict.update(
                {
                    (
                        el.get("host"),
                        el.get("community"),
                        el.get("version"),
                        inf,
                        tr,
                        el.get("udp_port"),
                    ): el,
                },
            )
        return host_dict

    def list_to_dict(self, config):
        data = deepcopy(config)

        if data.get("vrfs"):
            for x in data["vrfs"]:
                if "context" in x:
                    x["context"] = {y: {"name": y} for y in x["context"]}
                if "hosts" in x:
                    x["hosts"] = self._host_list_to_dict(x)

        pkey = {
            "communities": "name",
            "community_maps": "name",
            "interfaces": "name",
            "mib_schema": "name",
            "groups": "group",
            "mib_bulkstat_transfer_ids": "name",
            "users": "user",
            "vrfs": "vrf",
        }
        for k in pkey.keys():
            if k in data:
                data[k] = {i[pkey[k]]: i for i in data[k]}

        if "correlator" in data:
            if "rules" in data["correlator"]:
                data["correlator"]["rules"] = {
                    x["rule_name"]: x for x in data["correlator"]["rules"]
                }
            if "rule_sets" in data["correlator"]:
                data["correlator"]["rule_sets"] = {
                    x["name"]: x for x in data["correlator"]["rule_sets"]
                }

        if "context" in data:
            data["context"] = {x: {"name": x} for x in data["context"]}
        if "mib_object_lists" in data:
            data["mib_object_lists"] = {x: {"mib_object": x} for x in data["mib_object_lists"]}
        if "targets" in data:
            data["targets"] = {
                x["name"] + x.get("vrf", "") + x.get("host", ""): x for x in data["targets"]
            }
        if "hosts" in data:
            data["hosts"] = self._host_list_to_dict(data)
        return data



from __future__ import absolute_import, division, print_function

__metaclass__ = type

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.clickhouse.plugins.module_utils.clickhouse import (
    execute_query
)


class EntitySettings():
    def __init__(self, module: AnsibleModule, client, entity_name: str, entity_type: str):
        self.module = module
        self.client = client
        self.name = entity_name
        self.entity_type = entity_type.lower()

        if self.entity_type not in ("user", "role"):
            self.module.fail_json(msg=f"entity_type must be 'user' or 'role', got {entity_type}")

    def _get_where_column(self):
        return "user_name" if self.entity_type == "user" else "role_name"

    def fetch(self):
        """Fetch current settings from system.settings_profile_elements."""
        """Returns raw result."""
        column = self._get_where_column()
        query = (
            "SELECT setting_name, value, min, max, writability, inherit_profile "
            f"FROM system.settings_profile_elements "
            f"WHERE {column} = '{self.name}'"
        )

        result = execute_query(self.module, self.client, query)
        return result

    def _normalize_current_settings(self, rows):
        """Convert DB rows into easy-to-compare structures."""
        current_settings = {}
        current_profiles = []

        for row in rows:
            setting_name, value, min, max, writability, inherit_profile = row
            if inherit_profile:
                current_profiles.append(inherit_profile)
            else:
                current_settings[setting_name] = {}
                current_settings[setting_name]['value'] = value
                if min:
                    current_settings[setting_name]['min'] = min
                if max:
                    current_settings[setting_name]['max'] = max
                if writability:
                    current_settings[setting_name]['writability'] = writability

        return current_settings, sorted(current_profiles)

    def compare_and_build_clause(self, desired_settings, desired_profiles):
        # In case only one is passed
        if not desired_settings:
            desired_settings = {}
        if not desired_profiles:
            desired_profiles = []
        current_rows = self.fetch()
        current_settings, current_profiles = self._normalize_current_settings(current_rows)
        desired_settings = self._normalize_settings(desired_settings)

        # Compare
        settings_changed = current_settings != desired_settings
        profiles_changed = set(current_profiles) != set(desired_profiles)
        changed = settings_changed or profiles_changed

        if not changed:
            return False, "", {}

        parts = []

        # PROFILE 'name'
        for profile in desired_profiles:
            parts.append(f"PROFILE '{profile}'")

        # SETTINGS key='val' [MIN 'min_val'] [MAX 'max_val'] [WRITABLE|CONST|CHANGEABLE_IN_READONLY]
        for key, setting in desired_settings.items():
            prepared = f"{key}='{setting['value']}'"
            if setting.get('min'):
                prepared += f" MIN '{setting['min']}'"
            if setting.get('max'):
                prepared += f" MAX '{setting['max']}'"
            if setting.get('writability'):
                prepared += f" {setting['writability']}"
            parts.append(f"{prepared}")

        if desired_profiles or desired_settings:
            clause = " SETTINGS " + ", ".join(parts) if parts else ""
        else:
            clause = " SETTINGS NONE"

        # Simple diff for Ansible output
        diff = {
            "before": {"settings": current_settings, "profiles": current_profiles},
            "after": {"settings": desired_settings, "profiles": desired_profiles}
        }

        return changed, clause, diff

    def _normalize_settings(self, settings):
        if not settings:
            return {}
        # Map with DB aliases
        writability_map = {
            'READONLY': 'CONST',
            'CONST': 'CONST'
        }
        normalized = {}

        for setting_name, setting_config in settings.items():
            normalized_config = setting_config.copy()

            # Convert numbers to strings (for all fields)
            for prop_name, prop_value in normalized_config.items():
                if isinstance(prop_value, (int, float)):
                    normalized_config[prop_name] = str(prop_value)

            # Writable uppercase and replace aliases
            if normalized_config.get('writability', ''):
                writability_value = normalized_config['writability'].upper()
                normalized_config['writability'] = writability_map.get(
                    writability_value,
                    writability_value
                )

            normalized[setting_name] = normalized_config

        return normalized



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_policies_assignments_by_client_info
short_description: Information module for organizations _policies _assignments _by
  _client
description:
  - Get all organizations _policies _assignments _by _client.
  - Get policies for all clients with policies.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
  - cisco.meraki.module_info_pagination
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  organizationId:
    description:
      - OrganizationId path parameter. Organization ID.
    type: str
  perPage:
    description:
      - PerPage query parameter. The number of entries per page returned. Acceptable
        range is 3 - 1000. Default is 50.
    type: int
  startingAfter:
    description:
      - >
        StartingAfter query parameter. A token used by the server to indicate the
        start of the page. Often this is a timestamp or an ID but it is not limited
        to those. This parameter should not be defined by client applications. The
        link for the first, last, prev, or next page in the HTTP Link header should
        define it.
    type: str
  endingBefore:
    description:
      - >
        EndingBefore query parameter. A token used by the server to indicate the end
        of the page. Often this is a timestamp or an ID but it is not limited to those.
        This parameter should not be defined by client applications. The link for
        the first, last, prev, or next page in the HTTP Link header should define
        it.
    type: str
  t0:
    description:
      - T0 query parameter. The beginning of the timespan for the data. The maximum
        lookback period is 31 days from today.
    type: str
  timespan:
    description:
      - >
        Timespan query parameter. The timespan for which the information will be fetched.
        If specifying timespan, do not specify parameter t0. The value must be in
        seconds and be less than or equal to 31 days. The default is 1 day.
    type: float
  includeUndetectedClients:
    description:
      - >
        IncludeUndetectedClients query parameter. Include provisioned clients that
        have not associated to the network. Default false.
    type: bool
  networkIds:
    description:
      - NetworkIds query parameter. Network Ids (minimum 1, maximum 30).
    elements: str
    type: list
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for organizations getOrganizationPoliciesAssignmentsByClient
    description: Complete reference of the getOrganizationPoliciesAssignmentsByClient
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-organization-policies-assignments-by-client
notes:
  - SDK Method used are
    organizations.Organizations.get_organization_policies_assignments_by_client,
  - Paths used are
    get /organizations/{organizationId}/policies/assignments/byClient,
"""

EXAMPLES = r"""
- name: Get all organizations _policies _assignments _by _client
  cisco.meraki.organizations_policies_assignments_by_client_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    perPage: 0
    startingAfter: string
    endingBefore: string
    t0: string
    timespan: 0
    includeUndetectedClients: true
    networkIds: []
    organizationId: string
    total_pages: -1
    direction: next
  register: result
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: list
  elements: dict
  sample: >
    [
      {
        "name": "string",
        "clientId": "string",
        "mac": "string",
        "networkId": "string",
        "assigned": [
          {
            "name": "string",
            "type": "string",
            "id": "string",
            "limitTo": [
              {
                "appliance": true,
                "ssids": [
                  {
                    "number": 0
                  }
                ]
              }
            ]
          }
        ]
      }
    ]
"""



#!/usr/bin/python
# Copyright (c) 2021-2022 Hewlett Packard Enterprise, Inc. All rights reserved.
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
module: ilo_redfish_command
short_description: Manages Out-Of-Band controllers using Redfish APIs
version_added: 6.6.0
description:
  - Builds Redfish URIs locally and sends them to remote OOB controllers to perform an action.
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
extends_documentation_fragment:
  - community.general._attributes
  - community.general._redfish
options:
  category:
    required: true
    description:
      - Category to execute on OOB controller.
    type: str
    choices: ['Systems']
  command:
    required: true
    description:
      - List of commands to execute on OOB controller.
    type: list
    elements: str
  baseuri:
    required: true
    description:
      - Base URI of OOB controller.
    type: str
  username:
    description:
      - Username for authenticating to iLO.
    type: str
  password:
    description:
      - Password for authenticating to iLO.
    type: str
  auth_token:
    description:
      - Security token for authenticating to iLO.
    type: str
  timeout:
    description:
      - Timeout in seconds for HTTP requests to iLO.
    default: 60
    type: int
  validate_certs:
    version_added: 10.6.0
  ca_path:
    version_added: 10.6.0
  ciphers:
    version_added: 10.6.0
author:
  - Varni H P (@varini-hp)
"""

EXAMPLES = r"""
- name: Wait for iLO Reboot Completion
  community.general.ilo_redfish_command:
    category: Systems
    command: WaitforiLORebootCompletion
    baseuri: "{{ baseuri }}"
    username: "{{ username }}"
    password: "{{ password }}"
"""

RETURN = r"""
ilo_redfish_command:
  description: Returns the status of the operation performed on the iLO.
  type: dict
  contains:
    WaitforiLORebootCompletion:
      description: Returns the output msg and whether the function executed successfully.
      type: dict
      contains:
        ret:
          description: Return V(true)/V(false) based on whether the operation was performed successfully.
          type: bool
        msg:
          description: Status of the operation performed on the iLO.
          type: str
  returned: always
"""

# More will be added as module features are expanded
CATEGORY_COMMANDS_ALL = {"Systems": ["WaitforiLORebootCompletion"]}

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._ilo_redfish_utils import iLORedfishUtils
from ansible_collections.community.general.plugins.module_utils._redfish_utils import REDFISH_COMMON_ARGUMENT_SPEC


def main():
    result = {}
    argument_spec = dict(
        category=dict(required=True, choices=list(CATEGORY_COMMANDS_ALL.keys())),
        command=dict(required=True, type="list", elements="str"),
        baseuri=dict(required=True),
        timeout=dict(type="int", default=60),
        username=dict(),
        password=dict(no_log=True),
        auth_token=dict(no_log=True),
    )
    argument_spec.update(REDFISH_COMMON_ARGUMENT_SPEC)
    module = AnsibleModule(
        argument_spec,
        required_together=[
            ("username", "password"),
        ],
        required_one_of=[
            ("username", "auth_token"),
        ],
        mutually_exclusive=[
            ("username", "auth_token"),
        ],
        supports_check_mode=False,
    )

    category = module.params["category"]
    command_list = module.params["command"]

    # admin credentials used for authentication
    creds = {"user": module.params["username"], "pswd": module.params["password"], "token": module.params["auth_token"]}

    timeout = module.params["timeout"]

    # Build root URI
    root_uri = f"https://{module.params['baseuri']}"
    rf_utils = iLORedfishUtils(creds, root_uri, timeout, module)

    # Check that Category is valid
    if category not in CATEGORY_COMMANDS_ALL:
        module.fail_json(msg=f"Invalid Category '{category}'. Valid Categories = {list(CATEGORY_COMMANDS_ALL.keys())}")

    # Check that all commands are valid
    for cmd in command_list:
        # Fail if even one command given is invalid
        if cmd not in CATEGORY_COMMANDS_ALL[category]:
            module.fail_json(msg=f"Invalid Command '{cmd}'. Valid Commands = {CATEGORY_COMMANDS_ALL[category]}")

    if category == "Systems":
        # execute only if we find a System resource

        result = rf_utils._find_systems_resource()
        if result["ret"] is False:
            module.fail_json(msg=result["msg"])

        for command in command_list:
            if command == "WaitforiLORebootCompletion":
                result[command] = rf_utils.wait_for_ilo_reboot_completion()

    # Return data back or fail with proper message
    if not result[command]["ret"]:
        module.fail_json(msg=result)

    changed = result[command].get("changed", False)
    module.exit_json(ilo_redfish_command=result, changed=changed)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2024, Akini Ross (@akinross) <akinross@cisco.com>
# Copyright: (c) 2023, Anvitha Jain (@anvitha-jain) <anvjain@cisco.com>
# Copyright: (c) 2018, Dag Wieers (@dagwieers) <dag@wieers.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_template_vrf
short_description: Manage VRFs in schema templates
description:
- Manage VRFs in schema templates on Cisco ACI Multi-Site.
author:
- Akini Ross (@akinross)
- Anvitha Jain (@anvitha-jain)
- Dag Wieers (@dagwieers)
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: true
  template:
    description:
    - The name of the template.
    type: str
    required: true
  vrf:
    description:
    - The name of the VRF to manage.
    type: str
    aliases: [ name ]
  display_name:
    description:
    - The name as displayed on the MSO web interface.
    type: str
  layer3_multicast:
    description:
    - Whether to enable L3 multicast.
    type: bool
  vzany:
    description:
    - Whether to enable vzAny.
    type: bool
  ip_data_plane_learning:
    description:
    - Whether IP data plane learning is enabled or disabled.
    - The APIC defaults to C(enabled) when unset during creation.
    type: str
    choices: [ disabled, enabled ]
  preferred_group:
    description:
    - Whether to enable preferred Endpoint Group.
    - The APIC defaults to C(false) when unset during creation.
    type: bool
  site_aware_policy_enforcement:
    description:
    - Whether to enable Site-aware Policy Enforcement Mode.
    - Enabling or Disabling Site-aware Policy Enforcement Mode will cause temporary traffic disruption.
    - Enabling Site-aware Policy Enforcement Mode will increase TCAM usage for existing contracts.
    - Site-aware Policy Enforcement needs to be enabled in the consumer and provider VRF when an inter-VRF contract is required.
    - Contract permit logging cannot be used when Site-aware Policy Enforcement Mode is enabled.
    type: bool
  bd_enforcement_status:
    description:
    - Whether to enable BD Enforcement Status.
    type: bool
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Add a new VRF
  cisco.mso.mso_schema_template_vrf:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    vrf: VRF 1
    state: present

- name: Remove an VRF
  cisco.mso.mso_schema_template_vrf:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    vrf: VRF1
    state: absent

- name: Query a specific VRFs
  cisco.mso.mso_schema_template_vrf:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    vrf: VRF1
    state: query
  register: query_result

- name: Query all VRFs
  cisco.mso.mso_schema_template_vrf:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    state: query
  register: query_result
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec
from ansible_collections.cisco.mso.plugins.module_utils.schema import MSOSchema


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str", required=True),
        template=dict(type="str", required=True),
        vrf=dict(type="str", aliases=["name"]),  # This parameter is not required for querying all objects
        display_name=dict(type="str"),
        layer3_multicast=dict(type="bool"),
        vzany=dict(type="bool"),
        preferred_group=dict(type="bool"),
        ip_data_plane_learning=dict(type="str", choices=["enabled", "disabled"]),
        site_aware_policy_enforcement=dict(type="bool"),
        bd_enforcement_status=dict(type="bool"),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["vrf"]],
            ["state", "present", ["vrf"]],
        ],
    )

    schema = module.params.get("schema")
    template = module.params.get("template").replace(" ", "")
    vrf = module.params.get("vrf")
    display_name = module.params.get("display_name")
    layer3_multicast = module.params.get("layer3_multicast")
    vzany = module.params.get("vzany")
    ip_data_plane_learning = module.params.get("ip_data_plane_learning")
    preferred_group = module.params.get("preferred_group")
    site_aware_policy_enforcement = module.params.get("site_aware_policy_enforcement")
    bd_enforcement_status = module.params.get("bd_enforcement_status")
    state = module.params.get("state")

    vrfs_path = "/templates/{0}/vrfs".format(template)

    mso = MSOModule(module)

    mso_schema = MSOSchema(mso, schema, template)
    mso_schema.set_template(template)

    if state == "query":
        if vrf:
            mso_schema.set_template_vrf(vrf)
            mso.existing = mso_schema.schema_objects.get("template_vrf").details
        else:
            mso.existing = mso_schema.schema_objects.get("template").details.get("vrfs", [])
        mso.exit_json()

    mso_schema.set_template_vrf(vrf, False)

    template_vrf = mso_schema.schema_objects.get("template_vrf")
    ops = []

    mso.previous = mso.existing = template_vrf.details if template_vrf else mso.existing

    if state == "absent":
        if mso.existing:
            mso.sent = mso.existing = {}
            ops.append(dict(op="remove", path="{0}/{1}".format(vrfs_path, template_vrf.details.get("name"))))

    elif state == "present":
        if display_name is None and not mso.existing:
            display_name = vrf

        payload = dict(
            name=vrf,
            displayName=display_name,
            l3MCast=layer3_multicast,
            vzAnyEnabled=vzany,
            preferredGroup=preferred_group,
            ipDataPlaneLearning=ip_data_plane_learning,
        )

        if site_aware_policy_enforcement is not None:
            payload["siteAwarePolicyEnforcementMode"] = site_aware_policy_enforcement
        if bd_enforcement_status is not None:
            payload["bdEnfStatus"] = bd_enforcement_status

        if template_vrf and template_vrf.details.get("vrfRef"):
            # Add vrfRef from the details to ensure idempotency succeeds
            payload["vrfRef"] = template_vrf.details["vrfRef"]

        mso.sanitize(payload, collate=True)

        if mso.existing:
            vrf_path = "{0}/{1}".format(vrfs_path, template_vrf.details.get("name"))
            if display_name is not None and display_name != mso.existing.get("displayName"):
                ops.append(dict(op="replace", path=vrf_path + "/displayName", value=display_name))
            if layer3_multicast is not None and layer3_multicast != mso.existing.get("l3MCast"):
                ops.append(dict(op="replace", path=vrf_path + "/l3MCast", value=layer3_multicast))
            if vzany is not None and vzany != mso.existing.get("vzAnyEnabled"):
                ops.append(dict(op="replace", path=vrf_path + "/vzAnyEnabled", value=vzany))
            if preferred_group is not None and preferred_group != mso.existing.get("preferredGroup"):
                ops.append(dict(op="replace", path=vrf_path + "/preferredGroup", value=preferred_group))
            if ip_data_plane_learning is not None and ip_data_plane_learning != mso.existing.get("ipDataPlaneLearning"):
                ops.append(dict(op="replace", path=vrf_path + "/ipDataPlaneLearning", value=ip_data_plane_learning))
            if site_aware_policy_enforcement is not None and site_aware_policy_enforcement != mso.existing.get("siteAwarePolicyEnforcementMode"):
                ops.append(dict(op="replace", path=vrf_path + "/siteAwarePolicyEnforcementMode", value=site_aware_policy_enforcement))
            if bd_enforcement_status is not None and bd_enforcement_status != mso.existing.get("bdEnfStatus"):
                ops.append(dict(op="replace", path=vrf_path + "/bdEnfStatus", value=bd_enforcement_status))
        else:
            ops.append(dict(op="add", path=vrfs_path + "/-", value=mso.sent))

        mso.existing = mso.proposed

    if not module.check_mode and ops:
        mso.request(mso_schema.path, method="PATCH", data=ops)

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2017, Daniel Korn <korndaniel1@gmail.com>
# Copyright (c) 2017, Yaacov Zamir <yzamir@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: manageiq_tags

short_description: Management of resource tags in ManageIQ
extends_documentation_fragment:
  - community.general._manageiq
  - community.general._attributes

author: Daniel Korn (@dkorn)
description:
  - The manageiq_tags module supports adding, updating and deleting tags in ManageIQ.
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none

options:
  state:
    type: str
    description:
      - V(absent) - tags should not exist,
      - V(present) - tags should exist.
    choices: ['absent', 'present']
    default: 'present'
  tags:
    type: list
    elements: dict
    description:
      - V(tags) - list of dictionaries, each includes C(name) and C(category) keys.
      - Required if O(state) is V(present) or V(absent).
  resource_type:
    type: str
    description:
      - The relevant resource type in manageiq.
    required: true
    choices:
      - provider
      - host
      - vm
      - blueprint
      - category
      - cluster
      - data store
      - group
      - resource pool
      - service
      - service template
      - template
      - tenant
      - user
  resource_name:
    type: str
    description:
      - The name of the resource at which tags are be controlled.
      - Must be specified if O(resource_id) is not set. Both options are mutually exclusive.
  resource_id:
    description:
      - The ID of the resource at which tags are controlled.
      - Must be specified if O(resource_name) is not set. Both options are mutually exclusive.
    type: int
    version_added: 2.2.0
"""

EXAMPLES = r"""
- name: Create new tags for a provider in ManageIQ.
  community.general.manageiq_tags:
    resource_name: 'EngLab'
    resource_type: 'provider'
    tags:
      - category: environment
        name: prod
      - category: owner
        name: prod_ops
    manageiq_connection:
      url: 'http://127.0.0.1:3000'
      username: 'admin'
      password: 'smartvm'
      validate_certs: false # only do this when connecting to localhost!

- name: Create new tags for a provider in ManageIQ.
  community.general.manageiq_tags:
    resource_id: 23000000790497
    resource_type: 'provider'
    tags:
      - category: environment
        name: prod
      - category: owner
        name: prod_ops
    manageiq_connection:
      url: 'http://127.0.0.1:3000'
      username: 'admin'
      password: 'smartvm'
      validate_certs: false # only do this when connecting to localhost!

- name: Remove tags for a provider in ManageIQ.
  community.general.manageiq_tags:
    state: absent
    resource_name: 'EngLab'
    resource_type: 'provider'
    tags:
      - category: environment
        name: prod
      - category: owner
        name: prod_ops
    manageiq_connection:
      url: 'http://127.0.0.1:3000'
      username: 'admin'
      password: 'smartvm'
      validate_certs: false # only do this when connecting to localhost!
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._manageiq import (
    ManageIQ,
    ManageIQTags,
    manageiq_argument_spec,
    manageiq_entities,
)


def main():
    actions = {"present": "assign", "absent": "unassign"}
    argument_spec = dict(
        tags=dict(type="list", elements="dict"),
        resource_id=dict(type="int"),
        resource_name=dict(type="str"),
        resource_type=dict(required=True, type="str", choices=list(manageiq_entities().keys())),
        state=dict(type="str", choices=["present", "absent"], default="present"),
    )
    # add the manageiq connection arguments to the arguments
    argument_spec.update(manageiq_argument_spec())

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=[["resource_id", "resource_name"]],
        required_one_of=[["resource_id", "resource_name"]],
        required_if=[("state", "present", ["tags"]), ("state", "absent", ["tags"])],
    )

    tags = module.params["tags"]
    resource_id = module.params["resource_id"]
    resource_type_key = module.params["resource_type"]
    resource_name = module.params["resource_name"]
    state = module.params["state"]

    # get the action and resource type
    action = actions[state]
    resource_type = manageiq_entities()[resource_type_key]

    manageiq = ManageIQ(module)

    # query resource id, fail if resource does not exist
    if resource_id is None:
        resource_id = manageiq.query_resource_id(resource_type, resource_name)

    manageiq_tags = ManageIQTags(manageiq, resource_type, resource_id)

    # assign or unassign the tags
    res_args = manageiq_tags.assign_or_unassign_tags(tags, action)

    module.exit_json(**res_args)


if __name__ == "__main__":
    main()



"""
Generate MySQL sha256 compatible plugins hash for a given password and salt

based on
 * https://www.akkadia.org/drepper/SHA-crypt.txt
 * https://crypto.stackexchange.com/questions/77427/whats-the-algorithm-behind-mysqls-sha256-password-hashing-scheme/111174#111174
 * https://github.com/hashcat/hashcat/blob/master/tools/test_modules/m07400.pm
"""

from __future__ import absolute_import, division, print_function

__metaclass__ = type

import hashlib


def _to64(v, n):
    """Convert a 32-bit integer to a base-64 string"""
    i64 = (
        [".", "/"]
        + [chr(x) for x in range(48, 58)]
        + [chr(x) for x in range(65, 91)]
        + [chr(x) for x in range(97, 123)]
    )
    result = ""
    while n > 0:
        n -= 1
        result += i64[v & 0x3F]
        v >>= 6
    return result


def _hashlib_sha256(data):
    """Return SHA-256 digest from hashlib ."""
    return hashlib.sha256(data).digest()


def _sha256_digest(key, salt, loops):
    """Return a SHA-256 digest of the concatenation of the key, the salt, and the key, repeated as necessary."""
    # https://www.akkadia.org/drepper/SHA-crypt.txt
    num_bytes = 32
    bytes_key = key.encode()
    bytes_salt = salt.encode()
    digest_b = _hashlib_sha256(bytes_key + bytes_salt + bytes_key)

    tmp = bytes_key + bytes_salt
    for i in range(len(bytes_key), 0, -num_bytes):
        tmp += digest_b if i > num_bytes else digest_b[:i]

    i = len(bytes_key)
    while i > 0:
        tmp += digest_b if (i & 1) != 0 else bytes_key
        i >>= 1

    digest_a = _hashlib_sha256(tmp)

    tmp = b""
    for i in range(len(bytes_key)):
        tmp += bytes_key

    digest_dp = _hashlib_sha256(tmp)

    byte_sequence_p = b""
    for i in range(len(bytes_key), 0, -num_bytes):
        byte_sequence_p += digest_dp if i > num_bytes else digest_dp[:i]

    tmp = b""
    til = 16 + digest_a[0]

    for i in range(til):
        tmp += bytes_salt

    digest_ds = _hashlib_sha256(tmp)

    byte_sequence_s = b""
    for i in range(len(bytes_salt), 0, -num_bytes):
        byte_sequence_s += digest_ds if i > num_bytes else digest_ds[:i]

    digest_c = digest_a

    for i in range(loops):
        tmp = byte_sequence_p if (i & 1) else digest_c
        if i % 3:
            tmp += byte_sequence_s
        if i % 7:
            tmp += byte_sequence_p
        tmp += digest_c if (i & 1) else byte_sequence_p
        digest_c = _hashlib_sha256(tmp)

    inc1, inc2, mod, end = (10, 21, 30, 0)

    i = 0
    tmp = ""

    while True:
        tmp += _to64(
            (digest_c[i] << 16)
            | (digest_c[(i + inc1) % mod] << 8)
            | digest_c[(i + inc1 * 2) % mod],
            4,
        )
        i = (i + inc2) % mod
        if i == end:
            break

    tmp += _to64((digest_c[31] << 8) | digest_c[30], 3)

    return tmp


def mysql_sha256_password_hash(password, salt):
    """Return a MySQL compatible caching_sha2_password hash in raw format."""
    if len(salt) != 20:
        raise ValueError("Salt must be 20 characters long.")

    count = 5
    iteration = 1000 * count

    digest = _sha256_digest(password, salt, iteration)
    return "$A${0:>03}${1}{2}".format(count, salt, digest)


def mysql_sha256_password_hash_hex(password, salt):
    """Return a MySQL compatible caching_sha2_password hash in hex format."""
    return mysql_sha256_password_hash(password, salt).encode().hex().upper()



# (c) 2020, Alexei Znamensky <russoz@gmail.com>
# Copyright (c) 2020, Ansible Project
# Simplified BSD License (see LICENSES/BSD-2-Clause.txt or https://opensource.org/licenses/BSD-2-Clause)
# SPDX-License-Identifier: BSD-2-Clause

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import traceback
from contextlib import contextmanager
from functools import wraps

from ansible_collections.community.general.plugins.module_utils._mh.exceptions import (
    ModuleHelperException,
    _UnhandledSentinel,
)

_unhandled_exceptions: tuple[type[Exception], ...] = (_UnhandledSentinel,)


def cause_changes(when=None):
    def deco(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            try:
                func(self, *args, **kwargs)
                if when == "success":
                    self.changed = True
            except Exception:
                if when == "failure":
                    self.changed = True
                raise
            finally:
                if when == "always":
                    self.changed = True

        return wrapper

    return deco


@contextmanager
def no_handle_exceptions(*exceptions: type[Exception]):
    global _unhandled_exceptions
    current = _unhandled_exceptions
    _unhandled_exceptions = tuple(exceptions)
    try:
        yield
    finally:
        _unhandled_exceptions = current


def module_fails_on_exception(func):
    conflict_list = ("msg", "exception", "output", "vars", "changed")

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        def fix_key(k):
            return k if k not in conflict_list else f"_{k}"

        def fix_var_conflicts(output):
            result = {fix_key(k): v for k, v in output.items()}
            return result

        try:
            func(self, *args, **kwargs)
        except _unhandled_exceptions:
            # re-raise exception without further processing
            raise
        except ModuleHelperException as e:
            if e.update_output:
                self.update_output(e.update_output)
            # patchy solution to resolve conflict with output variables
            output = fix_var_conflicts(self.output)
            self.module.fail_json(
                msg=e.msg, exception=traceback.format_exc(), output=self.output, vars=self.vars.output(), **output
            )
        except Exception as e:
            # patchy solution to resolve conflict with output variables
            output = fix_var_conflicts(self.output)
            msg = f"Module failed with exception: {str(e).strip()}"
            self.module.fail_json(
                msg=msg, exception=traceback.format_exc(), output=self.output, vars=self.vars.output(), **output
            )

    return wrapper


def check_mode_skip(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self.module.check_mode:
            return func(self, *args, **kwargs)

    return wrapper


def check_mode_skip_returns(callable=None, value=None):
    def deco(func):
        if callable is not None:

            @wraps(func)
            def wrapper_callable(self, *args, **kwargs):
                if self.module.check_mode:
                    return callable(self, *args, **kwargs)
                return func(self, *args, **kwargs)

            return wrapper_callable

        else:

            @wraps(func)
            def wrapper_value(self, *args, **kwargs):
                if self.module.check_mode:
                    return value
                return func(self, *args, **kwargs)

            return wrapper_value

    return deco



#!/usr/bin/python

# Copyright (c) 2017, Eike Frost <ei@kefro.st>
# Copyright (c) 2021, Christophe Gilles <christophe.gilles54@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: keycloak_realm

short_description: Allows administration of Keycloak realm using Keycloak API

version_added: 3.0.0

description:
  - This module allows the administration of Keycloak realm using the Keycloak REST API. It requires access to the REST API
    using OpenID Connect; the user connecting and the realm being used must have the requisite access rights. In a default
    Keycloak installation, admin-cli and an admin user would work, as would a separate realm definition with the scope tailored
    to your needs and a user having the expected roles.
  - The names of module options are snake_cased versions of the camelCase ones found in the Keycloak API and its documentation
    at U(https://www.keycloak.org/docs-api/8.0/rest-api/index.html). Aliases are provided so camelCased versions can be used
    as well.
  - The Keycloak API does not always sanity check inputs, for example you can set SAML-specific settings on an OpenID Connect
    client for instance and also the other way around. B(Be careful). If you do not specify a setting, usually a sensible
    default is chosen.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
  action_group:
    version_added: 10.2.0

options:
  state:
    description:
      - State of the realm.
      - On V(present), the realm is created (or updated if it exists already).
      - On V(absent), the realm is removed if it exists.
    choices: ['present', 'absent']
    default: 'present'
    type: str

  id:
    description:
      - The realm to create.
    type: str
  realm:
    description:
      - The realm name.
    type: str
  access_code_lifespan:
    description:
      - The realm access code lifespan.
    aliases:
      - accessCodeLifespan
    type: int
  access_code_lifespan_login:
    description:
      - The realm access code lifespan login.
    aliases:
      - accessCodeLifespanLogin
    type: int
  access_code_lifespan_user_action:
    description:
      - The realm access code lifespan user action.
    aliases:
      - accessCodeLifespanUserAction
    type: int
  access_token_lifespan:
    description:
      - The realm access token lifespan.
    aliases:
      - accessTokenLifespan
    type: int
  access_token_lifespan_for_implicit_flow:
    description:
      - The realm access token lifespan for implicit flow.
    aliases:
      - accessTokenLifespanForImplicitFlow
    type: int
  account_theme:
    description:
      - The realm account theme.
    aliases:
      - accountTheme
    type: str
  action_token_generated_by_admin_lifespan:
    description:
      - The realm action token generated by admin lifespan.
    aliases:
      - actionTokenGeneratedByAdminLifespan
    type: int
  action_token_generated_by_user_lifespan:
    description:
      - The realm action token generated by user lifespan.
    aliases:
      - actionTokenGeneratedByUserLifespan
    type: int
  admin_events_details_enabled:
    description:
      - The realm admin events details enabled.
    aliases:
      - adminEventsDetailsEnabled
    type: bool
  admin_events_enabled:
    description:
      - The realm admin events enabled.
    aliases:
      - adminEventsEnabled
    type: bool
  admin_permissions_enabled:
    description:
      - The realm admin permissions enabled.
    aliases:
      - adminPermissionsEnabled
    type: bool
    version_added: 12.0.0
  admin_theme:
    description:
      - The realm admin theme.
    aliases:
      - adminTheme
    type: str
  attributes:
    description:
      - The realm attributes.
    type: dict
  browser_flow:
    description:
      - The realm browser flow.
    aliases:
      - browserFlow
    type: str
  browser_security_headers:
    description:
      - The realm browser security headers.
    aliases:
      - browserSecurityHeaders
    type: dict
  brute_force_protected:
    description:
      - The realm brute force protected.
    aliases:
      - bruteForceProtected
    type: bool
  brute_force_strategy:
    description:
      - The realm brute force strategy.
    aliases:
      - bruteForceStrategy
    choices: ['LINEAR', 'MULTIPLE']
    type: str
    version_added: 11.2.0
  client_authentication_flow:
    description:
      - The realm client authentication flow.
    aliases:
      - clientAuthenticationFlow
    type: str
  client_scope_mappings:
    description:
      - The realm client scope mappings.
    aliases:
      - clientScopeMappings
    type: dict
  default_default_client_scopes:
    description:
      - The realm default default client scopes.
    aliases:
      - defaultDefaultClientScopes
    type: list
    elements: str
  default_groups:
    description:
      - The realm default groups.
    aliases:
      - defaultGroups
    type: list
    elements: str
  default_locale:
    description:
      - The realm default locale.
    aliases:
      - defaultLocale
    type: str
  default_optional_client_scopes:
    description:
      - The realm default optional client scopes.
    aliases:
      - defaultOptionalClientScopes
    type: list
    elements: str
  default_roles:
    description:
      - The realm default roles.
    aliases:
      - defaultRoles
    type: list
    elements: str
  default_signature_algorithm:
    description:
      - The realm default signature algorithm.
    aliases:
      - defaultSignatureAlgorithm
    type: str
  direct_grant_flow:
    description:
      - The realm direct grant flow.
    aliases:
      - directGrantFlow
    type: str
  display_name:
    description:
      - The realm display name.
    aliases:
      - displayName
    type: str
  display_name_html:
    description:
      - The realm display name HTML.
    aliases:
      - displayNameHtml
    type: str
  docker_authentication_flow:
    description:
      - The realm docker authentication flow.
    aliases:
      - dockerAuthenticationFlow
    type: str
  duplicate_emails_allowed:
    description:
      - The realm duplicate emails allowed option.
    aliases:
      - duplicateEmailsAllowed
    type: bool
  edit_username_allowed:
    description:
      - The realm edit username allowed option.
    aliases:
      - editUsernameAllowed
    type: bool
  email_theme:
    description:
      - The realm email theme.
    aliases:
      - emailTheme
    type: str
  enabled:
    description:
      - The realm enabled option.
    type: bool
  enabled_event_types:
    description:
      - The realm enabled event types.
    aliases:
      - enabledEventTypes
    type: list
    elements: str
  events_enabled:
    description:
      - Enables or disables login events for this realm.
    aliases:
      - eventsEnabled
    type: bool
    version_added: 3.6.0
  events_expiration:
    description:
      - The realm events expiration.
    aliases:
      - eventsExpiration
    type: int
  events_listeners:
    description:
      - The realm events listeners.
    aliases:
      - eventsListeners
    type: list
    elements: str
  failure_factor:
    description:
      - The realm failure factor.
    aliases:
      - failureFactor
    type: int
  first_broker_login_flow:
    description:
      - The realm first broker login flow.
    aliases:
      - firstBrokerLoginFlow
    type: str
    version_added: 12.5.0
  internationalization_enabled:
    description:
      - The realm internationalization enabled option.
    aliases:
      - internationalizationEnabled
    type: bool
  localization_texts:
    description:
      - The custom localization texts for a realm.
    aliases:
      - localizationTexts
    type: dict
    version_added: 12.4.0
  login_theme:
    description:
      - The realm login theme.
    aliases:
      - loginTheme
    type: str
  login_with_email_allowed:
    description:
      - The realm login with email allowed option.
    aliases:
      - loginWithEmailAllowed
    type: bool
  max_delta_time_seconds:
    description:
      - The realm max delta time in seconds.
    aliases:
      - maxDeltaTimeSeconds
    type: int
  max_failure_wait_seconds:
    description:
      - The realm max failure wait in seconds.
    aliases:
      - maxFailureWaitSeconds
    type: int
  max_secondary_auth_failures:
    description:
      - The realm max secondary authentication failures (used with brute force detection for secondary auth mechanisms).
    aliases:
      - maxSecondaryAuthFailures
    type: int
    version_added: 13.1.0
  max_temporary_lockouts:
    description:
      - The realm max temporary lockouts.
    aliases:
      - maxTemporaryLockouts
    type: int
    version_added: 11.2.0
  minimum_quick_login_wait_seconds:
    description:
      - The realm minimum quick login wait in seconds.
    aliases:
      - minimumQuickLoginWaitSeconds
    type: int
  not_before:
    description:
      - The realm not before.
    aliases:
      - notBefore
    type: int
  offline_session_idle_timeout:
    description:
      - The realm offline session idle timeout.
    aliases:
      - offlineSessionIdleTimeout
    type: int
  offline_session_max_lifespan:
    description:
      - The realm offline session max lifespan.
    aliases:
      - offlineSessionMaxLifespan
    type: int
  offline_session_max_lifespan_enabled:
    description:
      - The realm offline session max lifespan enabled option.
    aliases:
      - offlineSessionMaxLifespanEnabled
    type: bool
  otp_policy_algorithm:
    description:
      - The realm otp policy algorithm.
    aliases:
      - otpPolicyAlgorithm
    type: str
  otp_policy_digits:
    description:
      - The realm otp policy digits.
    aliases:
      - otpPolicyDigits
    type: int
  otp_policy_initial_counter:
    description:
      - The realm otp policy initial counter.
    aliases:
      - otpPolicyInitialCounter
    type: int
  otp_policy_look_ahead_window:
    description:
      - The realm otp policy look ahead window.
    aliases:
      - otpPolicyLookAheadWindow
    type: int
  otp_policy_period:
    description:
      - The realm otp policy period.
    aliases:
      - otpPolicyPeriod
    type: int
  otp_policy_type:
    description:
      - The realm otp policy type.
    aliases:
      - otpPolicyType
    type: str
  otp_supported_applications:
    description:
      - The realm otp supported applications.
    aliases:
      - otpSupportedApplications
    type: list
    elements: str
  password_policy:
    description:
      - The realm password policy.
    aliases:
      - passwordPolicy
    type: str
  organizations_enabled:
    description:
      - Enables support for experimental organization feature.
    aliases:
      - organizationsEnabled
    type: bool
    version_added: 10.0.0
  permanent_lockout:
    description:
      - The realm permanent lockout.
    aliases:
      - permanentLockout
    type: bool
  quick_login_check_milli_seconds:
    description:
      - The realm quick login check in milliseconds.
    aliases:
      - quickLoginCheckMilliSeconds
    type: int
  refresh_token_max_reuse:
    description:
      - The realm refresh token max reuse.
    aliases:
      - refreshTokenMaxReuse
    type: int
  registration_allowed:
    description:
      - The realm registration allowed option.
    aliases:
      - registrationAllowed
    type: bool
  registration_email_as_username:
    description:
      - The realm registration email as username option.
    aliases:
      - registrationEmailAsUsername
    type: bool
  registration_flow:
    description:
      - The realm registration flow.
    aliases:
      - registrationFlow
    type: str
  remember_me:
    description:
      - The realm remember me option.
    aliases:
      - rememberMe
    type: bool
  reset_credentials_flow:
    description:
      - The realm reset credentials flow.
    aliases:
      - resetCredentialsFlow
    type: str
  reset_password_allowed:
    description:
      - The realm reset password allowed option.
    aliases:
      - resetPasswordAllowed
    type: bool
  revoke_refresh_token:
    description:
      - The realm revoke refresh token option.
    aliases:
      - revokeRefreshToken
    type: bool
  smtp_server:
    description:
      - The realm smtp server.
    aliases:
      - smtpServer
    type: dict
  ssl_required:
    description:
      - The realm ssl required option.
    choices: ['all', 'external', 'none']
    aliases:
      - sslRequired
    type: str
  sso_session_idle_timeout:
    description:
      - The realm sso session idle timeout.
    aliases:
      - ssoSessionIdleTimeout
    type: int
  sso_session_idle_timeout_remember_me:
    description:
      - The realm sso session idle timeout remember me.
    aliases:
      - ssoSessionIdleTimeoutRememberMe
    type: int
  sso_session_max_lifespan:
    description:
      - The realm sso session max lifespan.
    aliases:
      - ssoSessionMaxLifespan
    type: int
  sso_session_max_lifespan_remember_me:
    description:
      - The realm sso session max lifespan remember me.
    aliases:
      - ssoSessionMaxLifespanRememberMe
    type: int
  supported_locales:
    description:
      - The realm supported locales.
    aliases:
      - supportedLocales
    type: list
    elements: str
  user_managed_access_allowed:
    description:
      - The realm user managed access allowed option.
    aliases:
      - userManagedAccessAllowed
    type: bool
  verify_email:
    description:
      - The realm verify email option.
    aliases:
      - verifyEmail
    type: bool
  wait_increment_seconds:
    description:
      - The realm wait increment in seconds.
    aliases:
      - waitIncrementSeconds
    type: int
  client_session_idle_timeout:
    description:
      - All Clients inherit from this setting, time a session is allowed to be idle before it expires.
    aliases:
      - clientSessionIdleTimeout
    type: int
    version_added: 11.2.0
  client_session_max_lifespan:
    description:
      - All Clients inherit from this setting, max time before a session is expired.
    aliases:
      - clientSessionMaxLifespan
    type: int
    version_added: 11.2.0
  client_offline_session_idle_timeout:
    description:
      - All Clients inherit from this setting, time an offline session is allowed to be idle before it expires.
    aliases:
      - clientOfflineSessionIdleTimeout
    type: int
    version_added: 11.2.0
  client_offline_session_max_lifespan:
    description:
      - All Clients inherit from this setting, max time before an offline session is expired regardless of activity.
    aliases:
      - clientOfflineSessionMaxLifespan
    type: int
    version_added: 11.2.0
  oauth2_device_code_lifespan:
    description:
      - Max time before the device code and user code are expired.
    aliases:
      - oauth2DeviceCodeLifespan
    type: int
    version_added: 11.2.0
  oauth2_device_polling_interval:
    description:
      - The minimum amount of time in seconds that the client should wait between polling requests to the token endpoint.
    aliases:
      - oauth2DevicePollingInterval
    type: int
    version_added: 11.2.0
  web_authn_policy_rp_entity_name:
    description:
      - WebAuthn Relying Party Entity Name.
    aliases:
      - webAuthnPolicyRpEntityName
    type: str
    version_added: 11.3.0
  web_authn_policy_signature_algorithms:
    description:
      - List of acceptable WebAuthn signature algorithms.
    aliases:
      - webAuthnPolicySignatureAlgorithms
    type: list
    version_added: 11.3.0
    elements: str
  web_authn_policy_rp_id:
    description:
      - WebAuthn Relying Party ID (domain). Empty string means use request host.
    aliases:
      - webAuthnPolicyRpId
    type: str
    version_added: 11.3.0
  web_authn_policy_attestation_conveyance_preference:
    description:
      - Attestation conveyance preference for WebAuthn.
    aliases:
      - webAuthnPolicyAttestationConveyancePreference
    type: str
    version_added: 11.3.0
  web_authn_policy_authenticator_attachment:
    description:
      - Authenticator attachment preference for WebAuthn authenticators.
    aliases:
      - webAuthnPolicyAuthenticatorAttachment
    type: str
    version_added: 11.3.0
  web_authn_policy_require_resident_key:
    description:
      - Whether resident keys are required for WebAuthn (Yes/No/not specified).
    aliases:
      - webAuthnPolicyRequireResidentKey
    type: str
    version_added: 11.3.0
  web_authn_policy_user_verification_requirement:
    description:
      - User verification requirement for WebAuthn.
    aliases:
      - webAuthnPolicyUserVerificationRequirement
    type: str
    version_added: 11.3.0
  web_authn_policy_create_timeout:
    description:
      - Timeout for WebAuthn credential creation (ms).
    aliases:
      - webAuthnPolicyCreateTimeout
    type: int
    version_added: 11.3.0
  web_authn_policy_avoid_same_authenticator_register:
    description:
      - Avoid registering the same authenticator multiple times.
    aliases:
      - webAuthnPolicyAvoidSameAuthenticatorRegister
    type: bool
    version_added: 11.3.0
  web_authn_policy_acceptable_aaguids:
    description:
      - List of acceptable AAGUIDs for WebAuthn authenticators.
    aliases:
      - webAuthnPolicyAcceptableAaguids
    type: list
    version_added: 11.3.0
    elements: str
  web_authn_policy_extra_origins:
    description:
      - Additional acceptable origins for WebAuthn requests.
    aliases:
      - webAuthnPolicyExtraOrigins
    type: list
    version_added: 11.3.0
    elements: str
  web_authn_policy_passwordless_rp_entity_name:
    description:
      - WebAuthn Passwordless Relying Party Entity Name.
    aliases:
      - webAuthnPolicyPasswordlessRpEntityName
    type: str
    version_added: 11.3.0
  web_authn_policy_passwordless_signature_algorithms:
    description:
      - List of acceptable WebAuthn signature algorithms for passwordless.
    aliases:
      - webAuthnPolicyPasswordlessSignatureAlgorithms
    type: list
    version_added: 11.3.0
    elements: str
  web_authn_policy_passwordless_rp_id:
    description:
      - WebAuthn Passwordless Relying Party ID (domain).
    aliases:
      - webAuthnPolicyPasswordlessRpId
    type: str
    version_added: 11.3.0
  web_authn_policy_passwordless_attestation_conveyance_preference:
    description:
      - Attestation conveyance preference for WebAuthn passwordless.
    aliases:
      - webAuthnPolicyPasswordlessAttestationConveyancePreference
    type: str
    version_added: 11.3.0
  web_authn_policy_passwordless_authenticator_attachment:
    description:
      - Authenticator attachment for WebAuthn passwordless.
    aliases:
      - webAuthnPolicyPasswordlessAuthenticatorAttachment
    type: str
    version_added: 11.3.0
  web_authn_policy_passwordless_require_resident_key:
    description:
      - Whether resident keys are required for WebAuthn passwordless (V(Yes)/V(No)/V(not specified)).
    aliases:
      - webAuthnPolicyPasswordlessRequireResidentKey
    type: str
    version_added: 11.3.0
  web_authn_policy_passwordless_user_verification_requirement:
    description:
      - User verification requirement for WebAuthn passwordless.
    aliases:
      - webAuthnPolicyPasswordlessUserVerificationRequirement
    type: str
    version_added: 11.3.0
  web_authn_policy_passwordless_create_timeout:
    description:
      - Timeout for WebAuthn passwordless credential creation (ms).
    aliases:
      - webAuthnPolicyPasswordlessCreateTimeout
    type: int
    version_added: 11.3.0
  web_authn_policy_passwordless_avoid_same_authenticator_register:
    description:
      - Avoid registering the same authenticator multiple times for passwordless.
    aliases:
      - webAuthnPolicyPasswordlessAvoidSameAuthenticatorRegister
    type: bool
    version_added: 11.3.0
  web_authn_policy_passwordless_acceptable_aaguids:
    description:
      - List of acceptable AAGUIDs for WebAuthn passwordless authenticators.
    aliases:
      - webAuthnPolicyPasswordlessAcceptableAaguids
    type: list
    version_added: 11.3.0
    elements: str
  web_authn_policy_passwordless_extra_origins:
    description:
      - Additional acceptable origins for WebAuthn passwordless requests.
    aliases:
      - webAuthnPolicyPasswordlessExtraOrigins
    type: list
    version_added: 11.3.0
    elements: str
  web_authn_policy_passwordless_passkeys_enabled:
    description:
      - Enable passkeys (conditional UI) authentication in the username forms.
    aliases:
      - webAuthnPolicyPasswordlessPasskeysEnabled
    type: bool
    version_added: 12.1.0

extends_documentation_fragment:
  - community.general._keycloak
  - community.general._keycloak.actiongroup_keycloak
  - community.general._attributes

author:
  - Christophe Gilles (@kris2kris)
"""

EXAMPLES = r"""
- name: Create or update Keycloak realm (minimal example)
  community.general.keycloak_realm:
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
    realm: unique_realm_name
    state: present

- name: Delete a Keycloak realm
  community.general.keycloak_realm:
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
    realm: unique_realm_name
    state: absent
"""

RETURN = r"""
msg:
  description: Message as to what action was taken.
  returned: always
  type: str
  sample: "Realm testrealm has been updated"

proposed:
  description: Representation of proposed realm.
  returned: always
  type: dict
  sample: {"realm": "test"}

existing:
  description: Representation of existing realm (sample is truncated).
  returned: always
  type: dict
  sample:
    {
      "adminUrl": "http://www.example.com/admin_url",
      "attributes": {
        "request.object.signature.alg": "RS256"
      }
    }

end_state:
  description: Representation of realm after module execution (sample is truncated).
  returned: on success
  type: dict
  sample:
    {
      "adminUrl": "http://www.example.com/admin_url",
      "attributes": {
        "request.object.signature.alg": "RS256"
      }
    }
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._keycloak import (
    KeycloakAPI,
    KeycloakError,
    camel,
    get_token,
    keycloak_argument_spec,
)


def normalise_cr(realmrep):
    """Re-sorts any properties where the order is important so that diff's is minimised and the change detection is more effective.

    :param realmrep: the realmrep dict to be sanitized
    :return: normalised realmrep dict
    """
    # Avoid the dict passed in to be modified
    realmrep = realmrep.copy()

    if "enabledEventTypes" in realmrep:
        realmrep["enabledEventTypes"] = list(sorted(realmrep["enabledEventTypes"]))

    if "otpSupportedApplications" in realmrep:
        realmrep["otpSupportedApplications"] = list(sorted(realmrep["otpSupportedApplications"]))

    if "supportedLocales" in realmrep:
        realmrep["supportedLocales"] = list(sorted(realmrep["supportedLocales"]))

    return realmrep


def sanitize_cr(realmrep):
    """Removes probably sensitive details from a realm representation.

    :param realmrep: the realmrep dict to be sanitized
    :return: sanitized realmrep dict
    """
    if not realmrep:
        return realmrep

    result = realmrep.copy()
    if "secret" in result:
        result["secret"] = "********"
    if "attributes" in result:
        if "saml.signing.private.key" in result["attributes"]:
            result["attributes"] = result["attributes"].copy()
            result["attributes"]["saml.signing.private.key"] = "********"
    return normalise_cr(result)


def main():
    """
    Module execution

    :return:
    """
    argument_spec = keycloak_argument_spec()

    meta_args = dict(
        state=dict(default="present", choices=["present", "absent"]),
        id=dict(type="str"),
        realm=dict(type="str"),
        access_code_lifespan=dict(type="int", aliases=["accessCodeLifespan"]),
        access_code_lifespan_login=dict(type="int", aliases=["accessCodeLifespanLogin"]),
        access_code_lifespan_user_action=dict(type="int", aliases=["accessCodeLifespanUserAction"]),
        access_token_lifespan=dict(type="int", aliases=["accessTokenLifespan"], no_log=False),
        access_token_lifespan_for_implicit_flow=dict(
            type="int", aliases=["accessTokenLifespanForImplicitFlow"], no_log=False
        ),
        account_theme=dict(type="str", aliases=["accountTheme"]),
        action_token_generated_by_admin_lifespan=dict(
            type="int", aliases=["actionTokenGeneratedByAdminLifespan"], no_log=False
        ),
        action_token_generated_by_user_lifespan=dict(
            type="int", aliases=["actionTokenGeneratedByUserLifespan"], no_log=False
        ),
        admin_events_details_enabled=dict(type="bool", aliases=["adminEventsDetailsEnabled"]),
        admin_events_enabled=dict(type="bool", aliases=["adminEventsEnabled"]),
        admin_permissions_enabled=dict(type="bool", aliases=["adminPermissionsEnabled"]),
        admin_theme=dict(type="str", aliases=["adminTheme"]),
        attributes=dict(type="dict"),
        browser_flow=dict(type="str", aliases=["browserFlow"]),
        browser_security_headers=dict(type="dict", aliases=["browserSecurityHeaders"]),
        brute_force_protected=dict(type="bool", aliases=["bruteForceProtected"]),
        brute_force_strategy=dict(type="str", choices=["LINEAR", "MULTIPLE"], aliases=["bruteForceStrategy"]),
        client_authentication_flow=dict(type="str", aliases=["clientAuthenticationFlow"]),
        client_scope_mappings=dict(type="dict", aliases=["clientScopeMappings"]),
        default_default_client_scopes=dict(type="list", elements="str", aliases=["defaultDefaultClientScopes"]),
        default_groups=dict(type="list", elements="str", aliases=["defaultGroups"]),
        default_locale=dict(type="str", aliases=["defaultLocale"]),
        default_optional_client_scopes=dict(type="list", elements="str", aliases=["defaultOptionalClientScopes"]),
        default_roles=dict(type="list", elements="str", aliases=["defaultRoles"]),
        default_signature_algorithm=dict(type="str", aliases=["defaultSignatureAlgorithm"]),
        direct_grant_flow=dict(type="str", aliases=["directGrantFlow"]),
        display_name=dict(type="str", aliases=["displayName"]),
        display_name_html=dict(type="str", aliases=["displayNameHtml"]),
        docker_authentication_flow=dict(type="str", aliases=["dockerAuthenticationFlow"]),
        duplicate_emails_allowed=dict(type="bool", aliases=["duplicateEmailsAllowed"]),
        edit_username_allowed=dict(type="bool", aliases=["editUsernameAllowed"]),
        email_theme=dict(type="str", aliases=["emailTheme"]),
        enabled=dict(type="bool"),
        enabled_event_types=dict(type="list", elements="str", aliases=["enabledEventTypes"]),
        events_enabled=dict(type="bool", aliases=["eventsEnabled"]),
        events_expiration=dict(type="int", aliases=["eventsExpiration"]),
        events_listeners=dict(type="list", elements="str", aliases=["eventsListeners"]),
        failure_factor=dict(type="int", aliases=["failureFactor"]),
        first_broker_login_flow=dict(type="str", aliases=["firstBrokerLoginFlow"]),
        internationalization_enabled=dict(type="bool", aliases=["internationalizationEnabled"]),
        localization_texts=dict(type="dict", aliases=["localizationTexts"]),
        login_theme=dict(type="str", aliases=["loginTheme"]),
        login_with_email_allowed=dict(type="bool", aliases=["loginWithEmailAllowed"]),
        max_delta_time_seconds=dict(type="int", aliases=["maxDeltaTimeSeconds"]),
        max_failure_wait_seconds=dict(type="int", aliases=["maxFailureWaitSeconds"]),
        max_secondary_auth_failures=dict(type="int", aliases=["maxSecondaryAuthFailures"]),
        max_temporary_lockouts=dict(type="int", aliases=["maxTemporaryLockouts"]),
        minimum_quick_login_wait_seconds=dict(type="int", aliases=["minimumQuickLoginWaitSeconds"]),
        not_before=dict(type="int", aliases=["notBefore"]),
        offline_session_idle_timeout=dict(type="int", aliases=["offlineSessionIdleTimeout"]),
        offline_session_max_lifespan=dict(type="int", aliases=["offlineSessionMaxLifespan"]),
        offline_session_max_lifespan_enabled=dict(type="bool", aliases=["offlineSessionMaxLifespanEnabled"]),
        otp_policy_algorithm=dict(type="str", aliases=["otpPolicyAlgorithm"]),
        otp_policy_digits=dict(type="int", aliases=["otpPolicyDigits"]),
        otp_policy_initial_counter=dict(type="int", aliases=["otpPolicyInitialCounter"]),
        otp_policy_look_ahead_window=dict(type="int", aliases=["otpPolicyLookAheadWindow"]),
        otp_policy_period=dict(type="int", aliases=["otpPolicyPeriod"]),
        otp_policy_type=dict(type="str", aliases=["otpPolicyType"]),
        otp_supported_applications=dict(type="list", elements="str", aliases=["otpSupportedApplications"]),
        password_policy=dict(type="str", aliases=["passwordPolicy"], no_log=False),
        organizations_enabled=dict(type="bool", aliases=["organizationsEnabled"]),
        permanent_lockout=dict(type="bool", aliases=["permanentLockout"]),
        quick_login_check_milli_seconds=dict(type="int", aliases=["quickLoginCheckMilliSeconds"]),
        refresh_token_max_reuse=dict(type="int", aliases=["refreshTokenMaxReuse"], no_log=False),
        registration_allowed=dict(type="bool", aliases=["registrationAllowed"]),
        registration_email_as_username=dict(type="bool", aliases=["registrationEmailAsUsername"]),
        registration_flow=dict(type="str", aliases=["registrationFlow"]),
        remember_me=dict(type="bool", aliases=["rememberMe"]),
        reset_credentials_flow=dict(type="str", aliases=["resetCredentialsFlow"]),
        reset_password_allowed=dict(type="bool", aliases=["resetPasswordAllowed"], no_log=False),
        revoke_refresh_token=dict(type="bool", aliases=["revokeRefreshToken"]),
        smtp_server=dict(type="dict", aliases=["smtpServer"]),
        ssl_required=dict(choices=["external", "all", "none"], aliases=["sslRequired"]),
        sso_session_idle_timeout=dict(type="int", aliases=["ssoSessionIdleTimeout"]),
        sso_session_idle_timeout_remember_me=dict(type="int", aliases=["ssoSessionIdleTimeoutRememberMe"]),
        sso_session_max_lifespan=dict(type="int", aliases=["ssoSessionMaxLifespan"]),
        sso_session_max_lifespan_remember_me=dict(type="int", aliases=["ssoSessionMaxLifespanRememberMe"]),
        supported_locales=dict(type="list", elements="str", aliases=["supportedLocales"]),
        user_managed_access_allowed=dict(type="bool", aliases=["userManagedAccessAllowed"]),
        verify_email=dict(type="bool", aliases=["verifyEmail"]),
        wait_increment_seconds=dict(type="int", aliases=["waitIncrementSeconds"]),
        client_session_idle_timeout=dict(type="int", aliases=["clientSessionIdleTimeout"]),
        client_session_max_lifespan=dict(type="int", aliases=["clientSessionMaxLifespan"]),
        client_offline_session_idle_timeout=dict(type="int", aliases=["clientOfflineSessionIdleTimeout"]),
        client_offline_session_max_lifespan=dict(type="int", aliases=["clientOfflineSessionMaxLifespan"]),
        oauth2_device_code_lifespan=dict(type="int", aliases=["oauth2DeviceCodeLifespan"]),
        oauth2_device_polling_interval=dict(type="int", aliases=["oauth2DevicePollingInterval"]),
        web_authn_policy_rp_entity_name=dict(type="str", aliases=["webAuthnPolicyRpEntityName"]),
        web_authn_policy_signature_algorithms=dict(
            type="list", elements="str", aliases=["webAuthnPolicySignatureAlgorithms"]
        ),
        web_authn_policy_rp_id=dict(type="str", aliases=["webAuthnPolicyRpId"]),
        web_authn_policy_attestation_conveyance_preference=dict(
            type="str", aliases=["webAuthnPolicyAttestationConveyancePreference"]
        ),
        web_authn_policy_authenticator_attachment=dict(type="str", aliases=["webAuthnPolicyAuthenticatorAttachment"]),
        web_authn_policy_require_resident_key=dict(
            type="str", aliases=["webAuthnPolicyRequireResidentKey"], no_log=False
        ),
        web_authn_policy_user_verification_requirement=dict(
            type="str", aliases=["webAuthnPolicyUserVerificationRequirement"]
        ),
        web_authn_policy_create_timeout=dict(type="int", aliases=["webAuthnPolicyCreateTimeout"]),
        web_authn_policy_avoid_same_authenticator_register=dict(
            type="bool", aliases=["webAuthnPolicyAvoidSameAuthenticatorRegister"]
        ),
        web_authn_policy_acceptable_aaguids=dict(
            type="list", elements="str", aliases=["webAuthnPolicyAcceptableAaguids"]
        ),
        web_authn_policy_extra_origins=dict(type="list", elements="str", aliases=["webAuthnPolicyExtraOrigins"]),
        web_authn_policy_passwordless_rp_entity_name=dict(
            type="str", aliases=["webAuthnPolicyPasswordlessRpEntityName"]
        ),
        web_authn_policy_passwordless_signature_algorithms=dict(
            type="list", elements="str", aliases=["webAuthnPolicyPasswordlessSignatureAlgorithms"], no_log=False
        ),
        web_authn_policy_passwordless_rp_id=dict(type="str", aliases=["webAuthnPolicyPasswordlessRpId"]),
        web_authn_policy_passwordless_attestation_conveyance_preference=dict(
            type="str", aliases=["webAuthnPolicyPasswordlessAttestationConveyancePreference"], no_log=False
        ),
        web_authn_policy_passwordless_authenticator_attachment=dict(
            type="str", aliases=["webAuthnPolicyPasswordlessAuthenticatorAttachment"], no_log=False
        ),
        web_authn_policy_passwordless_require_resident_key=dict(
            type="str", aliases=["webAuthnPolicyPasswordlessRequireResidentKey"], no_log=False
        ),
        web_authn_policy_passwordless_user_verification_requirement=dict(
            type="str", aliases=["webAuthnPolicyPasswordlessUserVerificationRequirement"], no_log=False
        ),
        web_authn_policy_passwordless_create_timeout=dict(
            type="int", aliases=["webAuthnPolicyPasswordlessCreateTimeout"]
        ),
        web_authn_policy_passwordless_avoid_same_authenticator_register=dict(
            type="bool", aliases=["webAuthnPolicyPasswordlessAvoidSameAuthenticatorRegister"]
        ),
        web_authn_policy_passwordless_acceptable_aaguids=dict(
            type="list", elements="str", aliases=["webAuthnPolicyPasswordlessAcceptableAaguids"], no_log=False
        ),
        web_authn_policy_passwordless_extra_origins=dict(
            type="list", elements="str", aliases=["webAuthnPolicyPasswordlessExtraOrigins"], no_log=False
        ),
        web_authn_policy_passwordless_passkeys_enabled=dict(
            type="bool", aliases=["webAuthnPolicyPasswordlessPasskeysEnabled"]
        ),
    )

    argument_spec.update(meta_args)

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_one_of=(
            [
                ["id", "realm", "enabled"],
                ["token", "auth_realm", "auth_username", "auth_password", "auth_client_id", "auth_client_secret"],
            ]
        ),
        required_together=([["auth_username", "auth_password"]]),
        required_by={"refresh_token": "auth_realm"},
    )

    result = dict(changed=False, msg="", diff={}, proposed={}, existing={}, end_state={})

    # Obtain access token, initialize API
    try:
        connection_header = get_token(module.params)
    except KeycloakError as e:
        module.fail_json(msg=str(e))

    kc = KeycloakAPI(module, connection_header)

    realm = module.params.get("realm")
    state = module.params.get("state")

    # convert module parameters to realm representation parameters (if they belong in there)
    params_to_ignore = list(keycloak_argument_spec().keys()) + ["state"]

    # Filter and map the parameters names that apply to the role
    realm_params = [x for x in module.params if x not in params_to_ignore and module.params.get(x) is not None]

    # See whether the realm already exists in Keycloak
    before_realm = kc.get_realm_by_id(realm=realm)

    if before_realm is None:
        before_realm = {}

    # Build a proposed changeset from parameters given to this module
    changeset = {}

    for realm_param in realm_params:
        new_param_value = module.params.get(realm_param)
        changeset[camel(realm_param)] = new_param_value

    # Prepare the desired values using the existing values (non-existence results in a dict that is save to use as a basis)
    desired_realm = before_realm.copy()
    desired_realm.update(changeset)

    result["proposed"] = sanitize_cr(changeset)
    before_realm_sanitized = sanitize_cr(before_realm)
    result["existing"] = before_realm_sanitized

    # Cater for when it doesn't exist (an empty dict)
    if not before_realm:
        if state == "absent":
            # Do nothing and exit
            if module._diff:
                result["diff"] = dict(before="", after="")
            result["changed"] = False
            result["end_state"] = {}
            result["msg"] = "Realm does not exist, doing nothing."
            module.exit_json(**result)

        # Process a creation
        result["changed"] = True

        if module._diff:
            result["diff"] = dict(before="", after=sanitize_cr(desired_realm))

        if module.check_mode:
            module.exit_json(**result)

        # create it
        kc.create_realm(desired_realm)
        after_realm = kc.get_realm_by_id(desired_realm["realm"])

        result["end_state"] = sanitize_cr(after_realm)

        result["msg"] = f"Realm {desired_realm['realm']} has been created."
        module.exit_json(**result)

    else:
        if state == "present":
            # Process an update

            # doing an update
            result["changed"] = True
            if module.check_mode:
                # We can only compare the current realm with the proposed updates we have
                before_norm = normalise_cr(before_realm)
                desired_norm = normalise_cr(desired_realm)
                if module._diff:
                    result["diff"] = dict(before=sanitize_cr(before_norm), after=sanitize_cr(desired_norm))
                result["changed"] = before_norm != desired_norm

                module.exit_json(**result)

            # do the update
            kc.update_realm(desired_realm, realm=realm)

            after_realm = kc.get_realm_by_id(realm=realm)

            if before_realm == after_realm:
                result["changed"] = False

            result["end_state"] = sanitize_cr(after_realm)

            if module._diff:
                result["diff"] = dict(before=before_realm_sanitized, after=sanitize_cr(after_realm))

            result["msg"] = f"Realm {desired_realm['realm']} has been updated."
            module.exit_json(**result)

        else:
            # Process a deletion (because state was not 'present')
            result["changed"] = True

            if module._diff:
                result["diff"] = dict(before=before_realm_sanitized, after="")

            if module.check_mode:
                module.exit_json(**result)

            # delete it
            kc.delete_realm(realm=realm)

            result["proposed"] = {}
            result["end_state"] = {}

            result["msg"] = f"Realm {before_realm['realm']} has been deleted."

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_appliance_traffic_shaping_vpn_exclusions
short_description: Resource module for networks _appliance _traffic _shaping _vpn
  _exclusions
description:
  - Manage operation update of the resource networks _appliance _traffic _shaping
    _vpn _exclusions.
  - Update VPN exclusion rules for an MX network.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  custom:
    description: Custom VPN exclusion rules. Pass an empty array to clear existing
      rules.
    elements: dict
    suboptions:
      destination:
        description: Destination address; hostname required for DNS, IPv4 otherwise.
        type: str
      port:
        description: Destination port.
        type: str
      protocol:
        description: Protocol.
        type: str
    type: list
  majorApplications:
    description: Major Application based VPN exclusion rules. Pass an empty array
      to clear existing rules.
    elements: dict
    suboptions:
      id:
        description: Application's Meraki ID.
        type: str
      name:
        description: Application's name.
        type: str
    type: list
  networkId:
    description: NetworkId path parameter. Network ID.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for appliance updateNetworkApplianceTrafficShapingVpnExclusions
    description: Complete reference of the updateNetworkApplianceTrafficShapingVpnExclusions
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-network-appliance-traffic-shaping-vpn-exclusions
notes:
  - SDK Method used are
    appliance.Appliance.update_network_appliance_traffic_shaping_vpn_exclusions,
  - Paths used are
    put /networks/{networkId}/appliance/trafficShaping/vpnExclusions,
"""

EXAMPLES = r"""
- name: Update all
  cisco.meraki.networks_appliance_traffic_shaping_vpn_exclusions:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    custom:
      - destination: 192.168.3.0/24
        port: '8000'
        protocol: tcp
    majorApplications:
      - id: meraki:vpnExclusion/application/2
        name: Office 365 Sharepoint
    networkId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "networkId": "string",
      "networkName": "string",
      "custom": [
        {
          "protocol": "string",
          "destination": "string",
          "port": "string"
        }
      ],
      "majorApplications": [
        {
          "id": "string",
          "name": "string"
        }
      ]
    }
"""



# -*- coding: utf-8 -*-
# Copyright 2020 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The ios ospfv2 fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type


from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import utils
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)

from ansible_collections.cisco.ios.plugins.module_utils.network.ios.argspec.ospfv2.ospfv2 import (
    Ospfv2Args,
)
from ansible_collections.cisco.ios.plugins.module_utils.network.ios.rm_templates.ospfv2 import (
    Ospfv2Template,
)


class Ospfv2Facts(object):
    """The ios ospfv2 fact class"""

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = Ospfv2Args.argument_spec

    def get_ospfv2_data(self, connection):
        return connection.get("show running-config | section ^router ospf")

    def dict_to_list(self, ospf_data):
        """Converts areas, interfaces in each process to list
        :param ospf_data: ospf data
        :rtype: dictionary
        :returns: facts_output
        """

        facts_output = {"processes": []}

        for process in ospf_data.get("processes", []):
            if "passive_interfaces" in process and process["passive_interfaces"].get("default"):
                if process.get("passive_interfaces", {}).get("interface"):
                    process["passive_interfaces"]["interface"]["name"] = [
                        each for each in process["passive_interfaces"]["interface"]["name"] if each
                    ]
            if "areas" in process:
                process["areas"] = list(process["areas"].values())
            facts_output["processes"].append(process)

        return facts_output

    def populate_facts(self, connection, ansible_facts, data=None):
        """Populate the facts for ospfv2
        :param connection: the device connection
        :param ansible_facts: Facts dictionary
        :param data: previously collected conf
        :rtype: dictionary
        :returns: facts
        """

        facts = {}

        if not data:
            data = self.get_ospfv2_data(connection)

        ospf_temp_obj = NetworkTemplate(lines=data.splitlines(), tmplt=Ospfv2Template())
        ospf_parsed = ospf_temp_obj.parse()

        # Convert dict to list
        ospf_parsed["processes"] = (
            ospf_parsed["processes"].values() if "processes" in ospf_parsed else []
        )

        # converts areas, interfaces in each process to list
        facts_output = self.dict_to_list(ospf_parsed)

        ansible_facts["ansible_network_resources"].pop("ospfv2", None)

        if ospf_parsed["processes"]:
            params = utils.validate_config(self.argument_spec, {"config": facts_output})
            params = utils.remove_empties(params)
            facts["ospfv2"] = params["config"]
            ansible_facts["ansible_network_resources"].update(facts)
        return ansible_facts



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2015, Jose Armesto <jose@armesto.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: autoscaling_launch_config_find
version_added: 1.0.0
short_description: Find AWS Autoscaling Launch Configurations
description:
  - Returns list of matching Launch Configurations for a given name, along with other useful information.
  - Results can be sorted and sliced.
  - Based on the work by Tom Bamford U(https://github.com/tombamford).
  - Prior to release 5.0.0 this module was called C(community.aws.ec2_lc_find).
    The usage did not change.
author:
  - "Jose Armesto (@fiunchinho)"
options:
  name_regex:
    description:
      - A Launch Configuration to match.
      - It'll be compiled as regex.
    required: True
    type: str
  sort_order:
    description:
      - Order in which to sort results.
    choices: ['ascending', 'descending']
    default: 'ascending'
    type: str
  limit:
    description:
      - How many results to show.
      - Corresponds to Python slice notation like list[:limit].
    type: int
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.

- name: Search for the Launch Configurations that start with "app"
  community.aws.autoscaling_launch_config_find:
    name_regex: app.*
    sort_order: descending
    limit: 2
"""

RETURN = r"""
image_id:
    description: AMI id
    returned: when Launch Configuration was found
    type: str
    sample: "ami-0d75df7e"
user_data:
    description: User data used to start instance
    returned: when Launch Configuration was found
    type: str
    sample: "ZXhwb3J0IENMT1VE"
name:
    description: Name of the Launch Configuration
    returned: when Launch Configuration was found
    type: str
    sample: "myapp-v123"
arn:
    description: Name of the AMI
    returned: when Launch Configuration was found
    type: str
    sample: "arn:aws:autoscaling:eu-west-1:12345:launchConfiguration:d82f050e-e315:launchConfigurationName/yourproject"
instance_type:
    description: Type of ec2 instance
    returned: when Launch Configuration was found
    type: str
    sample: "t2.small"
created_time:
    description: When it was created
    returned: when Launch Configuration was found
    type: str
    sample: "2016-06-29T14:59:22.222000+00:00"
ebs_optimized:
    description: Launch Configuration EBS optimized property
    returned: when Launch Configuration was found
    type: bool
    sample: False
instance_monitoring:
    description: Launch Configuration instance monitoring property
    returned: when Launch Configuration was found
    type: str
    sample: {"Enabled": false}
classic_link_vpc_security_groups:
    description: Launch Configuration classic link vpc security groups property
    returned: when Launch Configuration was found
    type: list
    sample: []
block_device_mappings:
    description: Launch Configuration block device mappings property
    returned: when Launch Configuration was found
    type: list
    sample: []
keyname:
    description: Launch Configuration ssh key
    returned: when Launch Configuration was found
    type: str
    sample: mykey
security_groups:
    description: Launch Configuration security groups
    returned: when Launch Configuration was found
    type: list
    sample: []
kernel_id:
    description: Launch Configuration kernel to use
    returned: when Launch Configuration was found
    type: str
    sample: ''
ram_disk_id:
    description: Launch Configuration ram disk property
    returned: when Launch Configuration was found
    type: str
    sample: ''
associate_public_address:
    description: Assign public address or not
    returned: when Launch Configuration was found
    type: bool
    sample: True
...
"""

import re

try:
    import botocore
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


def find_launch_configs(client, module):
    name_regex = module.params.get("name_regex")
    sort_order = module.params.get("sort_order")
    limit = module.params.get("limit")

    paginator = client.get_paginator("describe_launch_configurations")

    response_iterator = paginator.paginate(PaginationConfig={"MaxItems": 1000, "PageSize": 100})

    results = []

    for response in response_iterator:
        response["LaunchConfigurations"] = filter(
            lambda lc: re.compile(name_regex).match(lc["LaunchConfigurationName"]), response["LaunchConfigurations"]
        )

        for lc in response["LaunchConfigurations"]:
            data = {
                "name": lc["LaunchConfigurationName"],
                "arn": lc["LaunchConfigurationARN"],
                "created_time": lc["CreatedTime"],
                "user_data": lc["UserData"],
                "instance_type": lc["InstanceType"],
                "image_id": lc["ImageId"],
                "ebs_optimized": lc["EbsOptimized"],
                "instance_monitoring": lc["InstanceMonitoring"],
                "classic_link_vpc_security_groups": lc["ClassicLinkVPCSecurityGroups"],
                "block_device_mappings": lc["BlockDeviceMappings"],
                "keyname": lc["KeyName"],
                "security_groups": lc["SecurityGroups"],
                "kernel_id": lc["KernelId"],
                "ram_disk_id": lc["RamdiskId"],
                "associate_public_address": lc.get("AssociatePublicIpAddress", False),
            }

            results.append(data)

    results.sort(key=lambda e: e["name"], reverse=(sort_order == "descending"))

    if limit:
        results = results[:int(limit)]  # fmt: skip

    module.exit_json(changed=False, results=results)


def main():
    argument_spec = dict(
        name_regex=dict(required=True),
        sort_order=dict(required=False, default="ascending", choices=["ascending", "descending"]),
        limit=dict(required=False, type="int"),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
    )

    try:
        client = module.client("autoscaling")
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Failed to connect to AWS")

    find_launch_configs(client, module)


if __name__ == "__main__":
    main()
